"""
When using GUI mode, this module's object (the main window) takes
user input to construct a program state (stored in the Core object).
This shows a preview of the video being created and allows for saving
projects and exporting the video at a later time.
"""

from PyQt6 import QtCore, QtWidgets, uic
import PyQt6.QtWidgets as QtWidgets
from PyQt6.QtGui import QShortcut
from PIL import Image
from queue import Queue
import sys
import os
import signal
import filecmp
import time
import logging
from textwrap import wrap

from ..__init__ import __version__
from ..core import Core, appName
from .undostack import UndoStack
from . import preview_thread
from .preview_win import PreviewWindow
from .presetmanager import PresetManager
from .actions import *
from ..toolkit.ffmpeg import createFfmpegCommand, checkFfmpegVersion
from ..toolkit import (
    disableWhenEncoding,
    disableWhenOpeningProject,
    checkOutput,
    blockSignals,
)


log = logging.getLogger("AVP.Gui.MainWindow")


class MainWindow(QtWidgets.QMainWindow):
    """
    The MainWindow wraps many Core methods in order to update the GUI
    accordingly. E.g., instead of self.core.openProject(), it will use
    self.openProject() and update the window titlebar within the wrapper.

    MainWindow manages the autosave feature, although Core has the
    primary functions for opening and creating project files.
    """

    createVideo = QtCore.pyqtSignal()
    newTask = QtCore.pyqtSignal(list)  # for the preview window
    processTask = QtCore.pyqtSignal()

    def __init__(self, project, dpi):
        super().__init__()
        log.debug("Main thread id: {}".format(int(QtCore.QThread.currentThreadId())))
        uic.loadUi(os.path.join(Core.wd, "gui", "mainwindow.ui"), self)

        if dpi:
            self.resize(
                int(self.width() * (dpi / 144)),
                int(self.height() * (dpi / 144)),
            )

        self.core = Core()
        Core.mode = "GUI"
        # widgets of component settings
        self.pages = []
        self.lastAutosave = time.time()
        # list of previous five autosave times, used to reduce update spam
        self.autosaveTimes = []
        self.autosaveCooldown = 0.2
        self.encoding = False

        # Find settings created by Core object
        self.dataDir = Core.dataDir
        self.presetDir = Core.presetDir
        self.autosavePath = os.path.join(self.dataDir, "autosave.avp")
        self.settings = Core.settings

        # Create stack of undoable user actions
        self.undoStack = UndoStack(self)
        undoLimit = self.settings.value("pref_undoLimit")
        self.undoStack.setUndoLimit(undoLimit)

        # Create Undo Dialog - A standard QUndoView on a standard QDialog
        self.undoDialog = QtWidgets.QDialog(self)
        self.undoDialog.setWindowTitle("Undo History")
        undoView = QtWidgets.QUndoView(self.undoStack)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(undoView)
        self.undoDialog.setLayout(layout)
        self.undoDialog.setMinimumWidth(int(self.width() / 2))

        # Create Preset Manager
        self.presetManager = PresetManager(self)

        # Create the preview window and its thread, queues, and timers
        log.debug("Creating preview window")
        self.previewWindow = PreviewWindow(
            self, os.path.join(Core.wd, "gui", "background.png")
        )
        self.verticalLayout_previewWrapper.addWidget(self.previewWindow)

        log.debug("Starting preview thread")
        self.previewQueue = Queue()
        self.previewThread = QtCore.QThread(self)
        self.previewWorker = preview_thread.Worker(
            self.core, self.settings, self.previewQueue
        )
        self.previewWorker.moveToThread(self.previewThread)
        self.newTask.connect(self.previewWorker.createPreviewImage)
        self.processTask.connect(self.previewWorker.process)
        self.previewWorker.error.connect(self.previewWindow.threadError)
        self.previewWorker.imageCreated.connect(self.showPreviewImage)
        self.previewThread.start()
        self.previewThread.finished.connect(
            lambda: log.info("Preview thread finished.")
        )

        timeout = 500
        log.debug("Preview timer set to trigger when idle for %sms" % str(timeout))
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.processTask.emit)
        self.timer.start(timeout)

        # Begin decorating the window and connecting events
        componentList = self.listWidget_componentList

        # Undo Feature
        def toggleUndoButtonEnabled(*_):
            """Enable/disable undo button depending on whether UndoStack contains Actions"""
            try:
                undoButton.setEnabled(self.undoStack.count())
            except RuntimeError:
                # program is probably in midst of exiting
                pass

        style = self.pushButton_undo.style()
        undoButton = self.pushButton_undo
        undoButton.setIcon(
            style.standardIcon(QtWidgets.QStyle.StandardPixmap.SP_FileDialogBack)
        )
        undoButton.clicked.connect(self.undoStack.undo)
        undoButton.setEnabled(False)
        self.undoStack.cleanChanged.connect(toggleUndoButtonEnabled)
        self.undoMenu = QtWidgets.QMenu()
        self.undoMenu.addAction(self.undoStack.createUndoAction(self))
        self.undoMenu.addAction(self.undoStack.createRedoAction(self))
        action = self.undoMenu.addAction("Show History...")
        action.triggered.connect(lambda _: self.showUndoStack())
        undoButton.setMenu(self.undoMenu)
        # end of Undo Feature

        style = self.pushButton_listMoveUp.style()
        self.pushButton_listMoveUp.setIcon(
            style.standardIcon(QtWidgets.QStyle.StandardPixmap.SP_ArrowUp)
        )
        style = self.pushButton_listMoveDown.style()
        self.pushButton_listMoveDown.setIcon(
            style.standardIcon(QtWidgets.QStyle.StandardPixmap.SP_ArrowDown)
        )
        style = self.pushButton_removeComponent.style()
        self.pushButton_removeComponent.setIcon(
            style.standardIcon(QtWidgets.QStyle.StandardPixmap.SP_DialogDiscardButton)
        )

        if sys.platform == "darwin":
            log.debug("Darwin detected: showing progress label below progress bar")
            self.progressBar_createVideo.setTextVisible(False)
        else:
            self.progressLabel.setHidden(True)

        self.toolButton_selectAudioFile.clicked.connect(self.openInputFileDialog)

        self.toolButton_selectOutputFile.clicked.connect(self.openOutputFileDialog)

        def changedField():
            self.autosave()
            self.updateWindowTitle()

        self.lineEdit_audioFile.textChanged.connect(changedField)
        self.lineEdit_outputFile.textChanged.connect(changedField)

        self.progressBar_createVideo.setValue(0)

        self.pushButton_createVideo.clicked.connect(self.createAudioVisualization)

        self.pushButton_Cancel.clicked.connect(self.stopVideo)

        for i, container in enumerate(Core.encoderOptions["containers"]):
            self.comboBox_videoContainer.addItem(container["name"])
            if container["name"] == self.settings.value("outputContainer"):
                selectedContainer = i

        self.comboBox_videoContainer.setCurrentIndex(selectedContainer)
        self.comboBox_videoContainer.currentIndexChanged.connect(self.updateCodecs)

        self.updateCodecs()

        for i in range(self.comboBox_videoCodec.count()):
            codec = self.comboBox_videoCodec.itemText(i)
            if codec == self.settings.value("outputVideoCodec"):
                self.comboBox_videoCodec.setCurrentIndex(i)

        for i in range(self.comboBox_audioCodec.count()):
            codec = self.comboBox_audioCodec.itemText(i)
            if codec == self.settings.value("outputAudioCodec"):
                self.comboBox_audioCodec.setCurrentIndex(i)

        self.comboBox_videoCodec.currentIndexChanged.connect(self.updateCodecSettings)

        self.comboBox_audioCodec.currentIndexChanged.connect(self.updateCodecSettings)

        vBitrate = int(self.settings.value("outputVideoBitrate"))
        aBitrate = int(self.settings.value("outputAudioBitrate"))

        self.spinBox_vBitrate.setValue(vBitrate)
        self.spinBox_aBitrate.setValue(aBitrate)
        self.spinBox_vBitrate.valueChanged.connect(self.updateCodecSettings)
        self.spinBox_aBitrate.valueChanged.connect(self.updateCodecSettings)

        # Make component buttons
        self.compMenu = QtWidgets.QMenu()
        for i, comp in enumerate(self.core.modules):
            action = self.compMenu.addAction(comp.Component.name)
            action.triggered.connect(lambda _, item=i: self.addComponent(0, item))

        self.pushButton_addComponent.setMenu(self.compMenu)

        componentList.dropEvent = self.dragComponent
        componentList.itemSelectionChanged.connect(self.changeComponentWidget)
        componentList.itemSelectionChanged.connect(
            self.presetManager.clearPresetListSelection
        )
        self.pushButton_removeComponent.clicked.connect(lambda: self.removeComponent())

        componentList.setContextMenuPolicy(
            QtCore.Qt.ContextMenuPolicy.CustomContextMenu
        )
        componentList.customContextMenuRequested.connect(self.componentContextMenu)

        currentRes = (
            str(self.settings.value("outputWidth"))
            + "x"
            + str(self.settings.value("outputHeight"))
        )
        for i, res in enumerate(Core.resolutions):
            self.comboBox_resolution.addItem(res)
            if res == currentRes:
                currentRes = i
                self.comboBox_resolution.setCurrentIndex(currentRes)
                self.comboBox_resolution.currentIndexChanged.connect(
                    self.updateResolution
                )

        self.pushButton_listMoveUp.clicked.connect(lambda: self.moveComponent(-1))
        self.pushButton_listMoveDown.clicked.connect(lambda: self.moveComponent(1))

        # Configure the Projects Menu
        self.projectMenu = QtWidgets.QMenu()
        self.menuButton_newProject = self.projectMenu.addAction("New Project")
        self.menuButton_newProject.triggered.connect(lambda: self.createNewProject())
        self.menuButton_openProject = self.projectMenu.addAction("Open Project")
        self.menuButton_openProject.triggered.connect(
            lambda: self.openOpenProjectDialog()
        )

        action = self.projectMenu.addAction("Save Project")
        action.triggered.connect(self.saveCurrentProject)

        action = self.projectMenu.addAction("Save Project As")
        action.triggered.connect(self.openSaveProjectDialog)

        self.pushButton_projects.setMenu(self.projectMenu)

        # Configure the Presets Button
        self.pushButton_presets.clicked.connect(self.openPresetManager)

        self.updateWindowTitle()
        log.debug("Showing main window")
        self.show()

        if project and project != self.autosavePath:
            if not project.endswith(".avp"):
                project += ".avp"
            # open a project from the commandline
            if not os.path.dirname(project):
                project = os.path.join(self.settings.value("projectDir"), project)
            self.currentProject = project
            self.settings.setValue("currentProject", project)
            if os.path.exists(self.autosavePath):
                os.remove(self.autosavePath)
        else:
            # open the last currentProject from settings
            self.currentProject = self.settings.value("currentProject")

            # delete autosave if it's identical to this project
            if self.autosaveExists(identical=True):
                os.remove(self.autosavePath)

            if self.currentProject and os.path.exists(self.autosavePath):
                ch = self.showMessage(
                    msg="Restore unsaved changes in project '%s'?"
                    % os.path.basename(self.currentProject)[:-4],
                    showCancel=True,
                )
                if ch:
                    self.saveProjectChanges()
                else:
                    os.remove(self.autosavePath)

        self.openProject(self.currentProject, prompt=False)
        self.drawPreview(True)

        log.info("Pillow version %s", Image.__version__)
        log.info(
            "PyQt version %s (Qt version %s)",
            QtCore.PYQT_VERSION_STR,
            QtCore.QT_VERSION_STR,
        )

        # verify Ffmpeg version
        if not self.core.FFMPEG_BIN:
            self.showMessage(
                msg="FFmpeg could not be found. This is a critical error. "
                "Install FFmpeg, or download it and place the program executable "
                "in the same folder as this program.",
                icon="Critical",
            )
        else:
            if not self.settings.value("ffmpegMsgShown"):
                ffmpegGoodVersion, ffmpegVersionNum = checkFfmpegVersion()
                if not ffmpegGoodVersion:
                    self.showMessage(
                        msg="The version of FFmpeg ({ffmpegVersionNum}) is "
                        "not recognized. Some features may not work as expected."
                    )
                self.settings.setValue("ffmpegMsgShown", True)

        # Hotkeys for projects

        QShortcut("Ctrl+S", self, self.saveCurrentProject)
        QShortcut("Ctrl+A", self, self.openSaveProjectDialog)
        QShortcut("Ctrl+O", self, self.openOpenProjectDialog)
        QShortcut("Ctrl+N", self, self.createNewProject)

        # Hotkeys for undo/redo
        QShortcut("Ctrl+Z", self, self.undoStack.undo)
        QShortcut("Ctrl+Y", self, self.undoStack.redo)
        QShortcut("Ctrl+Shift+Z", self, self.undoStack.redo)

        # Hotkeys for component list
        for inskey in ("Ctrl+T", QtCore.Qt.Key.Key_Insert):
            QShortcut(
                inskey,
                self,
                activated=lambda: self.pushButton_addComponent.click(),
            )
        for delkey in ("Ctrl+R", QtCore.Qt.Key.Key_Delete):
            QShortcut(delkey, self.listWidget_componentList, self.removeComponent)
        QShortcut(
            "Ctrl+Space",
            self,
            activated=lambda: self.listWidget_componentList.setFocus(),
        )
        QShortcut("Ctrl+Shift+S", self, self.presetManager.openSavePresetDialog)
        QShortcut("Ctrl+Shift+C", self, self.presetManager.clearPreset)

        QShortcut(
            "Ctrl+Up",
            self.listWidget_componentList,
            activated=lambda: self.moveComponent(-1),
        )
        QShortcut(
            "Ctrl+Down",
            self.listWidget_componentList,
            activated=lambda: self.moveComponent(1),
        )
        QShortcut(
            "Ctrl+Home",
            self.listWidget_componentList,
            activated=lambda: self.moveComponent("top"),
        )
        QShortcut(
            "Ctrl+End",
            self.listWidget_componentList,
            activated=lambda: self.moveComponent("bottom"),
        )

        QShortcut("F1", self, self.showHelpWindow)
        QShortcut("Ctrl+Shift+F", self, self.showFfmpegCommand)
        QShortcut("Ctrl+Shift+U", self, self.showUndoStack)

        if log.isEnabledFor(logging.DEBUG):
            QShortcut("Ctrl+Alt+Shift+R", self, self.drawPreview)
            QShortcut("Ctrl+Alt+Shift+A", self, lambda: log.debug(repr(self)))

        # Close MainWindow when receiving Ctrl+C from terminal
        signal.signal(signal.SIGINT, lambda *args: self.close())

        # Add initial components if none are in the list
        if not self.core.selectedComponents:
            self.core.insertComponent(
                0, self.core.moduleIndexFor("Classic Visualizer"), self
            )
            self.core.insertComponent(1, self.core.moduleIndexFor("Color"), self)
            # set colors to white and black to match classic appearance of program
            self.core.selectedComponents[0].page.lineEdit_visColor.setText(
                "255,255,255"
            )
            self.core.selectedComponents[1].page.lineEdit_color1.setText("0,0,0")
            self.undoStack.clear()

    def __repr__(self):
        return (
            "%s\n"
            "\n%s\n"
            "#####\n"
            "Preview thread is %s\n"
            % (
                super().__repr__(),
                (
                    "core not initialized"
                    if not hasattr(self, "core")
                    else repr(self.core)
                ),
                (
                    "live"
                    if hasattr(self, "previewThread") and self.previewThread.isRunning()
                    else "dead"
                ),
            )
        )

    def closeEvent(self, event):
        log.info("Ending the preview thread")
        self.timer.stop()
        self.previewThread.quit()
        self.previewThread.wait()
        return super().closeEvent(event)

    @disableWhenOpeningProject
    def updateWindowTitle(self):
        log.debug("Setting main window's title")
        windowTitle = appName
        try:
            if self.currentProject:
                windowTitle += (
                    " - %s" % os.path.splitext(os.path.basename(self.currentProject))[0]
                )
            if self.autosaveExists(identical=False):
                windowTitle += "*"
        except AttributeError:
            pass
        log.verbose(f'Window title is "{windowTitle}"')
        self.setWindowTitle(windowTitle)

    @QtCore.pyqtSlot(int, dict)
    def updateComponentTitle(self, pos, presetStore=False):
        """
        Sets component title to modified or unmodified when given boolean.
        If given a preset dict, compares it against the component to
        determine if it is modified.
        A component with no preset is always unmodified.
        """
        if type(presetStore) is dict:
            name = presetStore["preset"]
            if name is None or name not in self.core.savedPresets:
                modified = False
            else:
                modified = presetStore != self.core.savedPresets[name]

        modified = bool(presetStore)
        if pos < 0:
            pos = len(self.core.selectedComponents) - 1
        name = self.core.selectedComponents[pos].name
        title = str(name)
        if self.core.selectedComponents[pos].currentPreset:
            title += " - %s" % self.core.selectedComponents[pos].currentPreset
            if modified:
                title += "*"
        if type(presetStore) is bool:
            log.debug(
                "Forcing %s #%s's modified status to %s: %s",
                name,
                pos,
                modified,
                title,
            )
        else:
            log.debug("Setting %s #%s's title: %s", name, pos, title)
        self.listWidget_componentList.item(pos).setText(title)

    def updateCodecs(self):
        containerWidget = self.comboBox_videoContainer
        vCodecWidget = self.comboBox_videoCodec
        aCodecWidget = self.comboBox_audioCodec
        index = containerWidget.currentIndex()
        name = containerWidget.itemText(index)
        self.settings.setValue("outputContainer", name)

        vCodecWidget.clear()
        aCodecWidget.clear()

        for container in Core.encoderOptions["containers"]:
            if container["name"] == name:
                for vCodec in container["video-codecs"]:
                    vCodecWidget.addItem(vCodec)
                for aCodec in container["audio-codecs"]:
                    aCodecWidget.addItem(aCodec)

    def updateCodecSettings(self):
        """Updates settings.ini to match encoder option widgets"""
        vCodecWidget = self.comboBox_videoCodec
        vBitrateWidget = self.spinBox_vBitrate
        aBitrateWidget = self.spinBox_aBitrate
        aCodecWidget = self.comboBox_audioCodec
        currentVideoCodec = vCodecWidget.currentIndex()
        currentVideoCodec = vCodecWidget.itemText(currentVideoCodec)
        currentVideoBitrate = vBitrateWidget.value()
        currentAudioCodec = aCodecWidget.currentIndex()
        currentAudioCodec = aCodecWidget.itemText(currentAudioCodec)
        currentAudioBitrate = aBitrateWidget.value()
        self.settings.setValue("outputVideoCodec", currentVideoCodec)
        self.settings.setValue("outputAudioCodec", currentAudioCodec)
        self.settings.setValue("outputVideoBitrate", currentVideoBitrate)
        self.settings.setValue("outputAudioBitrate", currentAudioBitrate)

    @disableWhenOpeningProject
    def autosave(self, force=False):
        if not self.currentProject:
            if os.path.exists(self.autosavePath):
                os.remove(self.autosavePath)
        elif force or time.time() - self.lastAutosave >= self.autosaveCooldown:
            self.core.createProjectFile(self.autosavePath, self)
            self.lastAutosave = time.time()
            if len(self.autosaveTimes) >= 5:
                # Do some math to reduce autosave spam. This gives a smooth
                # curve up to 5 seconds cooldown and maintains that for 30 secs
                # if a component is continuously updated
                timeDiff = self.lastAutosave - self.autosaveTimes.pop()
                if not force and timeDiff >= 1.0 and timeDiff <= 10.0:
                    if self.autosaveCooldown / 4.0 < 0.5:
                        self.autosaveCooldown += 1.0
                    self.autosaveCooldown = (5.0 * (self.autosaveCooldown / 5.0)) + (
                        self.autosaveCooldown / 5.0
                    ) * 2
                elif force or timeDiff >= self.autosaveCooldown * 5:
                    self.autosaveCooldown = 0.2
            self.autosaveTimes.insert(0, self.lastAutosave)
        else:
            log.debug("Autosave rejected by cooldown")

    def autosaveExists(self, identical=True):
        """Determines if creating the autosave should be blocked."""
        try:
            if (
                self.currentProject
                and os.path.exists(self.autosavePath)
                and filecmp.cmp(self.autosavePath, self.currentProject) == identical
            ):
                log.debug(
                    "Autosave found %s to be identical" % "not" if not identical else ""
                )
                return True
        except FileNotFoundError:
            log.error("Project file couldn't be located: %s", self.currentProject)
            return identical
        return False

    def saveProjectChanges(self):
        """Overwrites project file with autosave file"""
        try:
            os.remove(self.currentProject)
            os.rename(self.autosavePath, self.currentProject)
            return True
        except (FileNotFoundError, IsADirectoryError) as e:
            self.showMessage(msg="Project file couldn't be saved.", detail=str(e))
            return False

    def openInputFileDialog(self):
        inputDir = self.settings.value("inputDir", os.path.expanduser("~"))

        fileName, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Open Audio File",
            inputDir,
            "Audio Files (%s)" % " ".join(Core.audioFormats),
        )

        if fileName:
            self.settings.setValue("inputDir", os.path.dirname(fileName))
            self.lineEdit_audioFile.setText(fileName)

    def openOutputFileDialog(self):
        outputDir = self.settings.value("outputDir", os.path.expanduser("~"))

        fileName, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Set Output Video File",
            outputDir,
            "Video Files (%s);; All Files (*)" % " ".join(Core.videoFormats),
        )

        if fileName:
            self.settings.setValue("outputDir", os.path.dirname(fileName))
            self.lineEdit_outputFile.setText(fileName)

    def stopVideo(self):
        log.info("Export cancelled")
        self.videoWorker.cancel()
        self.canceled = True

    def createAudioVisualization(self):
        # create output video if mandatory settings are filled in
        audioFile = self.lineEdit_audioFile.text()
        outputPath = self.lineEdit_outputFile.text()

        if audioFile and outputPath and self.core.selectedComponents:
            if not os.path.dirname(outputPath):
                outputPath = os.path.join(os.path.expanduser("~"), outputPath)
            if outputPath and os.path.isdir(outputPath):
                self.showMessage(
                    msg="Chosen filename matches a directory, which "
                    "cannot be overwritten. Please choose a different "
                    "filename or move the directory.",
                    icon="Warning",
                )
                return
        else:
            if not audioFile or not outputPath:
                self.showMessage(
                    msg="You must select an audio file and output filename."
                )
            elif not self.core.selectedComponents:
                self.showMessage(msg="Not enough components.")
            return

        self.canceled = False
        self.progressBarUpdated(-1)
        self.videoWorker = self.core.newVideoWorker(self, audioFile, outputPath)
        self.videoWorker.progressBarUpdate.connect(self.progressBarUpdated)
        self.videoWorker.progressBarSetText.connect(self.progressBarSetText)
        self.videoWorker.imageCreated.connect(self.showPreviewImage)
        self.videoWorker.encoding.connect(self.changeEncodingStatus)
        self.createVideo.emit()

    @QtCore.pyqtSlot(str, str)
    def videoThreadError(self, msg, detail):
        try:
            self.stopVideo()
        except AttributeError as e:
            if "videoWorker" not in str(e):
                raise
        self.showMessage(
            msg=msg,
            detail=detail,
            icon="Critical",
        )
        log.info("%s", repr(self))

    def changeEncodingStatus(self, status):
        self.encoding = status
        if status:
            # Disable many widgets when starting to export
            self.pushButton_createVideo.setEnabled(False)
            self.pushButton_Cancel.setEnabled(True)
            self.comboBox_resolution.setEnabled(False)
            self.stackedWidget.setEnabled(False)
            self.tab_encoderSettings.setEnabled(False)
            self.label_audioFile.setEnabled(False)
            self.toolButton_selectAudioFile.setEnabled(False)
            self.label_outputFile.setEnabled(False)
            self.toolButton_selectOutputFile.setEnabled(False)
            self.lineEdit_audioFile.setEnabled(False)
            self.lineEdit_outputFile.setEnabled(False)
            self.listWidget_componentList.setEnabled(False)
            self.pushButton_addComponent.setEnabled(False)
            self.pushButton_removeComponent.setEnabled(False)
            self.pushButton_listMoveDown.setEnabled(False)
            self.pushButton_listMoveUp.setEnabled(False)
            self.pushButton_undo.setEnabled(False)
            self.menuButton_newProject.setEnabled(False)
            self.menuButton_openProject.setEnabled(False)
            # Close undo history dialog if open
            self.undoDialog.close()
            # Show label under progress bar on macOS
            if sys.platform == "darwin":
                self.progressLabel.setHidden(False)
        else:
            self.pushButton_createVideo.setEnabled(True)
            self.pushButton_Cancel.setEnabled(False)
            self.comboBox_resolution.setEnabled(True)
            self.stackedWidget.setEnabled(True)
            self.tab_encoderSettings.setEnabled(True)
            self.label_audioFile.setEnabled(True)
            self.toolButton_selectAudioFile.setEnabled(True)
            self.lineEdit_audioFile.setEnabled(True)
            self.label_outputFile.setEnabled(True)
            self.toolButton_selectOutputFile.setEnabled(True)
            self.lineEdit_outputFile.setEnabled(True)
            self.pushButton_addComponent.setEnabled(True)
            self.pushButton_removeComponent.setEnabled(True)
            self.pushButton_listMoveDown.setEnabled(True)
            self.pushButton_listMoveUp.setEnabled(True)
            self.pushButton_undo.setEnabled(True)
            self.menuButton_newProject.setEnabled(True)
            self.menuButton_openProject.setEnabled(True)
            self.listWidget_componentList.setEnabled(True)
            self.progressLabel.setHidden(True)
            self.drawPreview(True)

    @QtCore.pyqtSlot(int)
    def progressBarUpdated(self, value):
        self.progressBar_createVideo.setValue(value)

    @QtCore.pyqtSlot(str)
    def progressBarSetText(self, value):
        if sys.platform == "darwin":
            self.progressLabel.setText(value)
        else:
            self.progressBar_createVideo.setFormat(value)
        if log.getEffectiveLevel() > logging.INFO:
            # if ffmpeg is quiet, print progress ourselves
            if any(
                [
                    value.startswith("Export C"),
                    value.startswith("Analyzing"),
                    value.startswith("Loading"),
                ]
            ):
                # Don't duplicate completion/failure messages or send too many messages
                return
            elif not value.startswith("Exporting"):
                print(value)
            else:
                # overwrite previous message with next one
                # if the text is our main export progress
                print(f"\r{value}", end="")

    def updateResolution(self):
        resIndex = int(self.comboBox_resolution.currentIndex())
        res = Core.resolutions[resIndex].split("x")
        changed = res[0] != self.settings.value("outputWidth")
        self.settings.setValue("outputWidth", res[0])
        self.settings.setValue("outputHeight", res[1])
        if changed:
            for i in range(len(self.core.selectedComponents)):
                self.core.updateComponent(i)

    def drawPreview(self, force=False, **kwargs):
        """Use autosave keyword arg to force saving or not saving if needed"""
        self.newTask.emit(self.core.selectedComponents)
        # self.processTask.emit()
        if force or "autosave" in kwargs:
            if force or kwargs["autosave"]:
                self.autosave(True)
        else:
            self.autosave()
        self.updateWindowTitle()

    @QtCore.pyqtSlot("QImage")
    def showPreviewImage(self, image):
        self.previewWindow.changePixmap(image)

    @disableWhenEncoding
    def showUndoStack(self):
        self.undoDialog.show()

    def showHelpWindow(self):
        self.showMessage(msg=f"{appName} v{__version__}")

    def showFfmpegCommand(self):
        command = createFfmpegCommand(
            self.lineEdit_audioFile.text(),
            self.lineEdit_outputFile.text(),
            self.core.selectedComponents,
        )
        command = " ".join(command)
        log.info(f"FFmpeg command: {command}")
        lines = wrap(command, 49)
        self.showMessage(msg=f"Current FFmpeg command:\n\n{' '.join(lines)}")

    def addComponent(self, compPos, moduleIndex):
        """Creates an undoable action that adds a new component."""
        action = AddComponent(self, compPos, moduleIndex)
        self.undoStack.push(action)

    def insertComponent(self, index):
        """Triggered by Core to finish initializing a new component."""
        if not hasattr(self.core.selectedComponents[index], "page"):
            log.error("Component failed to initialize")
            return
        componentList = self.listWidget_componentList
        stackedWidget = self.stackedWidget

        componentList.insertItem(index, self.core.selectedComponents[index].name)
        componentList.setCurrentRow(index)

        # connect to signal that adds an asterisk when modified
        self.core.selectedComponents[index].modified.connect(self.updateComponentTitle)

        self.pages.insert(index, self.core.selectedComponents[index].page)
        stackedWidget.insertWidget(index, self.pages[index])
        stackedWidget.setCurrentIndex(index)

        return index

    def removeComponent(self):
        componentList = self.listWidget_componentList
        selected = componentList.selectedItems()
        if selected:
            action = RemoveComponent(self, selected)
            self.undoStack.push(action)

    def _removeComponent(self, index):
        stackedWidget = self.stackedWidget
        componentList = self.listWidget_componentList
        stackedWidget.removeWidget(self.pages[index])
        componentList.takeItem(index)
        self.core.removeComponent(index)
        self.pages.pop(index)
        self.changeComponentWidget()
        self.drawPreview()

    @disableWhenEncoding
    def moveComponent(self, change):
        """Moves a component relatively from its current position"""
        componentList = self.listWidget_componentList
        tag = change
        if change == "top":
            change = -componentList.currentRow()
        elif change == "bottom":
            change = len(componentList) - componentList.currentRow() - 1
        else:
            tag = "down" if change == 1 else "up"

        row = componentList.currentRow()
        newRow = row + change
        if newRow > -1 and newRow < componentList.count():
            action = MoveComponent(self, row, newRow, tag)
            self.undoStack.push(action)

    def getComponentListMousePos(self, position):
        """
        Given a QPos, returns the component index under the mouse cursor
        or -1 if no component is there.
        """
        componentList = self.listWidget_componentList

        if hasattr(position, "toPointF"):
            position = position.toPointF()
        position = position.toPoint()

        modelIndexes = [
            componentList.model().index(i) for i in range(componentList.count())
        ]
        rects = [componentList.visualRect(modelIndex) for modelIndex in modelIndexes]
        mousePos = [rect.contains(position) for rect in rects]
        if not any(mousePos):
            # Not clicking a component
            mousePos = -1
        else:
            mousePos = mousePos.index(True)
        log.debug("Click component list row %s" % mousePos)
        return mousePos

    @disableWhenEncoding
    def dragComponent(self, event):
        """Used as Qt drop event for the component listwidget"""
        componentList = self.listWidget_componentList
        mousePos = self.getComponentListMousePos(event.position())

        if mousePos > -1:
            change = (componentList.currentRow() - mousePos) * -1
        else:
            change = componentList.count() - componentList.currentRow() - 1
        self.moveComponent(change)

    def changeComponentWidget(self):
        selected = self.listWidget_componentList.selectedItems()
        if selected:
            index = self.listWidget_componentList.row(selected[0])
            self.stackedWidget.setCurrentIndex(index)

    def openPresetManager(self):
        """Preset manager for importing, exporting, renaming, deleting"""
        self.presetManager.show_()

    def clear(self):
        """Get a blank slate"""
        self.core.clearComponents()
        self.listWidget_componentList.clear()
        for widget in self.pages:
            self.stackedWidget.removeWidget(widget)
        self.pages = []
        for field in (self.lineEdit_audioFile, self.lineEdit_outputFile):
            with blockSignals(field):
                field.setText("")
        self.progressBarUpdated(0)
        self.progressBarSetText("")
        self.undoStack.clear()

    @disableWhenEncoding
    def createNewProject(self, prompt=True):
        if prompt:
            ch = self.openSaveChangesDialog("starting a new project")
            if ch is None:
                return

        self.clear()
        self.currentProject = None
        self.settings.setValue("currentProject", None)
        self.drawPreview(True)

    def saveCurrentProject(self):
        if self.currentProject:
            self.core.createProjectFile(self.currentProject, self)
            try:
                os.remove(self.autosavePath)
            except FileNotFoundError:
                pass
            self.updateWindowTitle()
        else:
            self.openSaveProjectDialog()

    def openSaveChangesDialog(self, phrase):
        success = True
        ch = True
        if self.autosaveExists(identical=False):
            ch = self.showMessage(
                msg="You have unsaved changes in project '%s'. "
                "Save before %s?"
                % (os.path.basename(self.currentProject)[:-4], phrase),
                showDiscard=True,
            )
            if ch:
                success = self.saveProjectChanges()
        if ch is not None and success and os.path.exists(self.autosavePath):
            os.remove(self.autosavePath)
        return success and ch

    def openSaveProjectDialog(self):
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Create Project File",
            self.settings.value("projectDir"),
            "Project Files (*.avp)",
        )
        if not filename:
            return
        if not filename.endswith(".avp"):
            filename += ".avp"
        self.settings.setValue("projectDir", os.path.dirname(filename))
        self.settings.setValue("currentProject", filename)
        self.currentProject = filename
        self.core.createProjectFile(filename, self)
        self.updateWindowTitle()

    @disableWhenEncoding
    def openOpenProjectDialog(self):
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Open Project File",
            self.settings.value("projectDir"),
            "Project Files (*.avp)",
        )
        self.openProject(filename)

    def openProject(self, filepath, prompt=True):
        if (
            not filepath
            or not os.path.exists(filepath)
            or not filepath.endswith(".avp")
        ):
            return

        # ask to save any changes that are about to get deleted
        if prompt:
            ch = self.openSaveChangesDialog("opening another project")
            if ch is None:
                return
        self.clear()

        self.currentProject = filepath
        self.settings.setValue("currentProject", filepath)
        self.settings.setValue("projectDir", os.path.dirname(filepath))
        # actually load the project using core method
        self.core.openProject(self, filepath)
        self.drawPreview(autosave=False)
        self.updateWindowTitle()

    def showMessage(self, **kwargs):
        parent = kwargs["parent"] if "parent" in kwargs else self
        msg = QtWidgets.QMessageBox(parent)
        msg.setWindowTitle(appName)
        msg.setModal(True)
        msg.setText(kwargs["msg"])
        msg.setIcon(
            eval("QtWidgets.QMessageBox.Icon.%s" % kwargs["icon"])
            if "icon" in kwargs
            else QtWidgets.QMessageBox.Icon.Information
        )
        msg.setDetailedText(kwargs["detail"] if "detail" in kwargs else None)
        if "showDiscard" in kwargs and kwargs["showDiscard"]:
            msg.setStandardButtons(
                QtWidgets.QMessageBox.StandardButton.Save
                | QtWidgets.QMessageBox.StandardButton.Discard
                | QtWidgets.QMessageBox.StandardButton.Cancel
            )
        elif "showCancel" in kwargs and kwargs["showCancel"]:
            msg.setStandardButtons(
                QtWidgets.QMessageBox.StandardButton.Ok
                | QtWidgets.QMessageBox.StandardButton.Cancel
            )
        else:
            msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
        ch = msg.exec()
        if ch == 1024 or ch == 2048:
            # OK or Save
            return True
        elif ch > 8000000:
            # Discard
            return False
        # Cancel
        return None

    @disableWhenEncoding
    def componentContextMenu(self, QPos):
        """Appears when right-clicking the component list"""
        componentList = self.listWidget_componentList
        self.menu = QtWidgets.QMenu()
        parentPosition = componentList.mapToGlobal(QtCore.QPoint(0, 0))

        index = self.getComponentListMousePos(QPos)
        if index > -1:
            # Show preset menu if clicking a component
            self.presetManager.findPresets()
            menuItem = self.menu.addAction("Save Preset")
            menuItem.triggered.connect(self.presetManager.openSavePresetDialog)

            # submenu for opening presets
            try:
                presets = self.presetManager.presets[
                    str(self.core.selectedComponents[index])
                ]
                self.presetSubmenu = QtWidgets.QMenu("Open Preset")
                self.menu.addMenu(self.presetSubmenu)

                for version, presetName in presets:
                    menuItem = self.presetSubmenu.addAction(presetName)
                    menuItem.triggered.connect(
                        lambda _, presetName=presetName: self.presetManager.openPreset(
                            presetName
                        )
                    )
            except KeyError:
                pass

            if self.core.selectedComponents[index].currentPreset:
                menuItem = self.menu.addAction("Clear Preset")
                menuItem.triggered.connect(self.presetManager.clearPreset)
            self.menu.addSeparator()

        # "Add Component" submenu
        self.submenu = QtWidgets.QMenu("Add")
        self.menu.addMenu(self.submenu)
        insertCompAtTop = self.settings.value("pref_insertCompAtTop")
        for i, comp in enumerate(self.core.modules):
            menuItem = self.submenu.addAction(comp.Component.name)
            menuItem.triggered.connect(
                lambda _, item=i: self.addComponent(
                    0 if insertCompAtTop else index, item
                )
            )

        self.menu.move(parentPosition + QPos)
        self.menu.show()



"""Functions used to transform and manipulate audio for use by visualizers"""

from copy import copy
import numpy


def createSpectrumArray(
    component,
    completeAudioArray,
    sampleSize,
    smoothConstantDown,
    smoothConstantUp,
    scale,
    progressBarUpdate,
    progressBarSetText,
):
    lastProgress = 0
    lastSpectrum = None
    spectrumArray = {}
    for i in range(0, len(completeAudioArray), sampleSize):
        if component.canceled:
            break
        lastSpectrum = transformData(
            i,
            completeAudioArray,
            sampleSize,
            smoothConstantDown,
            smoothConstantUp,
            lastSpectrum,
            scale,
        )
        spectrumArray[i] = copy(lastSpectrum)

        progress = int(100 * (i / len(completeAudioArray)))
        if progress >= 100:
            progress = 100
        if progress == lastProgress:
            continue
        progressText = f"Analyzing audio: {str(progress)}%"
        progressBarSetText.emit(progressText)
        progressBarUpdate.emit(int(progress))
        lastProgress = progress
    return spectrumArray


def transformData(
    i,
    completeAudioArray,
    sampleSize,
    smoothConstantDown,
    smoothConstantUp,
    lastSpectrum,
    scale,
):
    if len(completeAudioArray) < (i + sampleSize):
        sampleSize = len(completeAudioArray) - i

    window = numpy.hanning(sampleSize)
    data = completeAudioArray[i : i + sampleSize][::1] * window
    paddedSampleSize = 2048
    paddedData = numpy.pad(data, (0, paddedSampleSize - sampleSize), "constant")
    spectrum = numpy.fft.fft(paddedData)
    sample_rate = 44100
    frequencies = numpy.fft.fftfreq(len(spectrum), 1.0 / sample_rate)

    y = abs(spectrum[0 : int(paddedSampleSize / 2) - 1])

    # filter the noise away
    # y[y<80] = 0

    with numpy.errstate(divide="ignore"):
        y = scale * numpy.log10(y)

    y[numpy.isinf(y)] = 0

    if lastSpectrum is not None:
        lastSpectrum[y < lastSpectrum] = y[
            y < lastSpectrum
        ] * smoothConstantDown + lastSpectrum[y < lastSpectrum] * (
            1 - smoothConstantDown
        )

        lastSpectrum[y >= lastSpectrum] = y[
            y >= lastSpectrum
        ] * smoothConstantUp + lastSpectrum[y >= lastSpectrum] * (1 - smoothConstantUp)
    else:
        lastSpectrum = y

    x = frequencies[0 : int(paddedSampleSize / 2) - 1]

    return lastSpectrum



"""
When using commandline mode, this module's object handles interpreting
the arguments and giving them to Core, which tracks the main program state.
Then it immediately exports a video.
"""

from PyQt6 import QtCore
import argparse
import os
import sys
import time
import glob
import signal
import shutil
import logging

from . import __version__
from .core import Core


log = logging.getLogger("AVP.Commandline")


class Command(QtCore.QObject):
    """
    This replaces the GUI MainWindow when in commandline mode.
    """

    createVideo = QtCore.pyqtSignal()

    def __init__(self):
        super().__init__()
        self.core = Core()
        Core.mode = "commandline"
        self.dataDir = self.core.dataDir
        self.canceled = False
        self.settings = Core.settings

        # ctrl-c stops the export thread
        signal.signal(signal.SIGINT, self.stopVideo)

    def parseArgs(self):
        parser = argparse.ArgumentParser(
            prog="avp" if os.path.basename(sys.argv[0]) == "__main__.py" else None,
            description="Create a visualization for an audio file",
            epilog="EXAMPLE COMMAND:   avp myvideotemplate "
            "-i ~/Music/song.mp3 -o ~/video.mp4 "
            "-c 0 image path=~/Pictures/thisWeeksPicture.jpg "
            '-c 1 video "preset=My Logo" -c 2 vis layout=classic',
        )

        parser.add_argument(
            "--version", "-V", action="version", version=f"%(prog)s {__version__}"
        )

        # input/output automatic-export commands
        parser.add_argument("-i", "--input", metavar="SOUND", help="input audio file")
        parser.add_argument(
            "-o", "--output", metavar="OUTPUT", help="output video file"
        )
        parser.add_argument(
            "--export-project",
            action="store_true",
            help="use input and output files from project file if -i or -o is missing",
        )

        # mutually exclusive debug options
        debugCommands = parser.add_mutually_exclusive_group()
        debugCommands.add_argument(
            "--log",
            action="store_true",
            help="copy and shorten recent log files into ~/avp_log.txt",
        )
        debugCommands.add_argument(
            "--verbose",
            "-v",
            action="store_true",
            help="send log messages and ffmpeg output to stdout, and create more verbose log files (good to use before --log)",
        )

        # project/GUI options
        parser.add_argument(
            "projpath",
            metavar="path-to-project",
            help="open a project file (.avp)",
            nargs="?",
        )
        parser.add_argument(
            "-c",
            "--comp",
            metavar=("LAYER", "ARG"),
            help="first arg must be component NAME to insert at LAYER."
            '"help" for information about possible args for a component.',
            nargs="*",
            action="append",
        )
        parser.add_argument(
            "--no-preview",
            action="store_true",
            help="disable live preview during export",
        )

        args = parser.parse_args()

        if args.verbose:
            Core.stdoutLogLvl = logging.DEBUG
            Core.makeLogger(deleteOldLogs=False, fileLogLvl=logging.DEBUG)

        if args.log:
            self.createLogFile()
            quit(0)

        if args.projpath:
            projPath = args.projpath
            if not os.path.dirname(projPath):
                projPath = os.path.join(self.settings.value("projectDir"), projPath)
            if not projPath.endswith(".avp"):
                projPath += ".avp"
            self.core.openProject(self, projPath)
            self.core.selectedComponents = list(reversed(self.core.selectedComponents))
            self.core.componentListChanged()

        if args.comp:
            for comp in args.comp:
                pos = comp[0]
                name = comp[1]
                compargs = comp[2:]
                try:
                    pos = int(pos)
                except ValueError:
                    print(pos, "is not a layer number.")
                    quit(1)
                realName = self.parseCompName(name)
                if not realName:
                    print(name, "is not a valid component name.")
                    quit(1)
                modI = self.core.moduleIndexFor(realName)
                i = self.core.insertComponent(pos, modI, self)
                if i is None:
                    print(name, "could not be initialized.")
                    quit(1)
                for arg in compargs:
                    self.core.selectedComponents[i].command(arg)

        if args.export_project and args.projpath:
            errcode, data = self.core.parseAvFile(projPath)
            input_ = None
            output = None
            for key, value in data["WindowFields"]:
                if "outputFile" in key:
                    output = value
                    if output and not os.path.dirname(value):
                        output = os.path.join(os.path.expanduser("~"), output)
                if "audioFile" in key:
                    input_ = value

            # use input/output from project file, overwritten by -i and -o
            if (not input_ and not args.input) or (not output and not args.output):
                parser.print_help()
                quit(1)

            self.createAudioVisualization(
                input_ if not args.input else args.input,
                output if not args.output else args.output,
            )
            return "commandline"

        elif args.input and args.output:
            self.createAudioVisualization(args.input, args.output)
            return "commandline"

        elif args.no_preview:
            Core.previewEnabled = False

        elif (
            args.projpath is None
            and "help" not in sys.argv
            and "--verbose" not in sys.argv
            and "-v" not in sys.argv
            and "--log" not in sys.argv
        ):
            parser.print_help()
            quit(1)

        return "GUI"

    def createAudioVisualization(self, input, output):
        if not self.core.selectedComponents:
            print("No components selected. Adding a default visualizer.")
            time.sleep(1)
            self.core.insertComponent(0, 0, self)
        self.core.selectedComponents = list(reversed(self.core.selectedComponents))
        self.core.componentListChanged()
        self.worker = self.core.newVideoWorker(self, input, output)
        # quit(0) after video is created
        self.worker.videoCreated.connect(self.videoCreated)
        self.lastProgressUpdate = time.time()
        self.worker.progressBarSetText.connect(self.progressBarSetText)
        self.createVideo.emit()

    def stopVideo(self, *args):
        self.worker.error = True
        self.worker.cancelExport()
        self.worker.cancel()

    @QtCore.pyqtSlot(str)
    def progressBarSetText(self, value):
        if "Export " in value or time.time() - self.lastProgressUpdate < 0.1:
            # Don't duplicate completion/failure messages or send too many messages
            return

        if not value.endswith("%"):
            # Show most messages very often
            print(value)
        elif log.getEffectiveLevel() > logging.INFO:
            # if ffmpeg isn't printing export progress for us,
            # then overwrite previous message with the next one
            # if this text is our main export progress
            print(f"{value}\r", end="")
        self.lastProgressUpdate = time.time()

    @QtCore.pyqtSlot()
    def videoCreated(self):
        self.quit(0)

    def quit(self, code):
        print()
        quit(code)

    def showMessage(self, **kwargs):
        print(kwargs["msg"])
        if "detail" in kwargs:
            print(kwargs["detail"])

    @QtCore.pyqtSlot(str, str)
    def videoThreadError(self, msg, detail):
        print(msg)
        print(detail)
        quit(1)

    def drawPreview(self, *args):
        pass

    def parseCompName(self, name):
        """Deduces a proper component name out of a commandline arg"""
        if name.title() in self.core.compNames:
            return name.title()
        for compName in self.core.compNames:
            if name.capitalize() in compName:
                return compName
        for altName, moduleIndex in self.core.altCompNames:
            if name.title() in altName:
                return self.core.compNames[moduleIndex]

        compFileNames = [
            os.path.splitext(os.path.basename(mod.__file__))[0]
            for mod in self.core.modules
        ]
        for i, compFileName in enumerate(compFileNames):
            if name.lower() in compFileName:
                return self.core.compNames[i]

        return None

    def createLogFile(self):
        # Choose a numbered location to put the output file
        logNumber = 0

        def getFilename():
            """Get a numbered filename for the final log file"""
            nonlocal logNumber
            name = os.path.join(os.path.expanduser("~"), "avp_log")
            while True:
                possibleName = f"{name}{logNumber:0>2}.txt"
                if os.path.exists(possibleName) and logNumber < 100:
                    logNumber += 1
                    continue
                break
            return possibleName

        # Copy latest debug log to chosen log file location
        filename = getFilename()
        if logNumber == 100:
            print("Log file could not be created (too many exist).")
            return
        try:
            shutil.copy(os.path.join(Core.logDir, "avp_debug.log"), filename)
            with open(filename, "a") as f:
                f.write(f"{'='*60} debug log ends {'='*60}\n")
        except FileNotFoundError:
            print("No debug log was found. Run `avp --verbose` before `avp --log`.")
            with open(filename, "w") as f:
                f.write(f"{'='*60} no debug log {'='*60}\n")

        def concatenateLogs(logPattern):
            nonlocal filename
            renderLogs = glob.glob(os.path.join(Core.logDir, logPattern))
            with open(filename, "a") as fw:
                for renderLog in renderLogs:
                    with open(renderLog, "r") as fr:
                        fw.write(f"{'='*60} {os.path.basename(renderLog)} {'='*60}\n")
                        logContents = fr.readlines()
                        fw.write("".join(logContents[:5]))
                        fw.write("...trimmed...\n")
                        fw.write("".join(logContents[-10:]))
                        fw.write(f"{'='*60} {os.path.basename(renderLog)} {'='*60}\n")

        concatenateLogs("render_*.log")
        concatenateLogs("preview_*.log")
        print(f"Log file created at {filename}")



import numpy
from avp.toolkit.frame import BlankFrame, FloodFrame


def test_blank_frame():
    """BlankFrame creates a frame of all zeros"""
    assert numpy.asarray(BlankFrame(1920, 1080), dtype="int32").sum() == 0


def test_flood_frame():
    """FloodFrame given (1, 1, 1, 1) creates a frame of sum 1920 * 1080 * 4"""
    assert numpy.asarray(FloodFrame(1920, 1080, (1, 1, 1, 1)), dtype="int32").sum() == (
        1920 * 1080 * 4
    )



from PIL import Image
import os
import subprocess
import logging

from ..libcomponent import BaseComponent
from ..toolkit.frame import BlankFrame, scale
from ..toolkit import connectWidget
from ..toolkit.ffmpeg import (
    openPipe,
    closePipe,
    getAudioDuration,
    FfmpegVideo,
    exampleSound,
)


log = logging.getLogger("AVP.Components.Spectrum")


class Component(BaseComponent):
    name = "Spectrum"
    version = "1.0.1"

    def widget(self, *args):
        self.previewFrame = None
        super().widget(*args)
        self._image = BlankFrame(self.width, self.height)
        self.chunkSize = 4 * self.width * self.height
        self.changedOptions = True
        self.previewSize = (214, 120)
        self.previewPipe = None

        if hasattr(self.parent, "lineEdit_audioFile"):
            # update preview when audio file changes (if genericPreview is off)
            self.parent.lineEdit_audioFile.textChanged.connect(self.update)

        self.trackWidgets(
            {
                "filterType": self.page.comboBox_filterType,
                "window": self.page.comboBox_window,
                "mode": self.page.comboBox_mode,
                "amplitude": self.page.comboBox_amplitude0,
                "amplitude1": self.page.comboBox_amplitude1,
                "amplitude2": self.page.comboBox_amplitude2,
                "display": self.page.comboBox_display,
                "zoom": self.page.spinBox_zoom,
                "tc": self.page.spinBox_tc,
                "x": self.page.spinBox_x,
                "y": self.page.spinBox_y,
                "mirror": self.page.checkBox_mirror,
                "draw": self.page.checkBox_draw,
                "scale": self.page.spinBox_scale,
                "color": self.page.comboBox_color,
                "compress": self.page.checkBox_compress,
                "mono": self.page.checkBox_mono,
                "hue": self.page.spinBox_hue,
            },
            relativeWidgets=[
                "x",
                "y",
            ],
        )
        for widget in self._trackedWidgets.values():
            connectWidget(widget, lambda: self.changed())

    def changed(self):
        self.changedOptions = True

    def update(self):
        filterType = self.page.comboBox_filterType.currentIndex()
        self.page.stackedWidget.setCurrentIndex(filterType)
        if filterType == 3:
            self.page.spinBox_hue.setEnabled(False)
        else:
            self.page.spinBox_hue.setEnabled(True)
        if filterType == 2 or filterType == 4:
            self.page.checkBox_mono.setEnabled(False)
        else:
            self.page.checkBox_mono.setEnabled(True)

    def previewRender(self):
        changedSize = self.updateChunksize()
        if (
            not changedSize
            and not self.changedOptions
            and self.previewFrame is not None
        ):
            log.debug("Spectrum #%s is reusing old preview frame" % self.compPos)
            return self.previewFrame

        frame = self.getPreviewFrame()
        self.changedOptions = False
        if not frame:
            log.warning("Spectrum #%s failed to create a preview frame" % self.compPos)
            self.previewFrame = None
            return BlankFrame(self.width, self.height)
        else:
            self.previewFrame = frame
            return frame

    def preFrameRender(self, **kwargs):
        super().preFrameRender(**kwargs)
        if self.previewPipe is not None:
            self.previewPipe.wait()
        self.updateChunksize()
        w, h = scale(self.scale, self.width, self.height, str)
        self.video = FfmpegVideo(
            inputPath=self.audioFile,
            filter_=self.makeFfmpegFilter(),
            width=w,
            height=h,
            chunkSize=self.chunkSize,
            frameRate=int(self.settings.value("outputFrameRate")),
            parent=self.parent,
            component=self,
        )

    def frameRender(self, frameNo):
        if FfmpegVideo.threadError is not None:
            raise FfmpegVideo.threadError
        return self.finalizeFrame(self.video.frame(frameNo))

    def postFrameRender(self):
        closePipe(self.video.pipe)

    def getPreviewFrame(self):
        genericPreview = self.settings.value("pref_genericPreview")
        startPt = 0
        if not genericPreview:
            inputFile = self.parent.lineEdit_audioFile.text()
            if not inputFile or not os.path.exists(inputFile):
                return
            duration = getAudioDuration(inputFile)
            if not duration:
                return
            startPt = duration / 3

        command = [
            self.core.FFMPEG_BIN,
            "-thread_queue_size",
            "512",
            "-r",
            str(self.settings.value("outputFrameRate")),
            "-ss",
            "{0:.3f}".format(startPt),
            "-i",
            self.core.junkStream if genericPreview else inputFile,
            "-f",
            "image2pipe",
            "-pix_fmt",
            "rgba",
        ]
        command.extend(self.makeFfmpegFilter(preview=True, startPt=startPt))
        command.extend(
            [
                "-an",
                "-s:v",
                "%sx%s" % scale(self.scale, self.width, self.height, str),
                "-codec:v",
                "rawvideo",
                "-",
                "-frames:v",
                "1",
            ]
        )

        if self.core.logEnabled:
            logFilename = os.path.join(
                self.core.logDir, "preview_%s.log" % str(self.compPos)
            )
            log.debug("Creating FFmpeg process (log at %s)" % logFilename)
            with open(logFilename, "w") as logf:
                logf.write(" ".join(command) + "\n\n")
            with open(logFilename, "a") as logf:
                self.previewPipe = openPipe(
                    command,
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.PIPE,
                    stderr=logf,
                    bufsize=10**8,
                )
        else:
            self.previewPipe = openPipe(
                command,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                bufsize=10**8,
            )
        byteFrame = self.previewPipe.stdout.read(self.chunkSize)
        closePipe(self.previewPipe)

        frame = self.finalizeFrame(byteFrame)
        return frame

    def makeFfmpegFilter(self, preview=False, startPt=0):
        """Makes final FFmpeg filter command"""

        def getFilterComplexCommand():
            """Inner function that creates the final, complex part of the filter command"""
            nonlocal self
            genericPreview = self.settings.value("pref_genericPreview")

            def getFilterComplexCommandForType():
                """Determine portion of filter command that changes depending on selected type"""
                nonlocal self
                if preview:
                    w, h = self.previewSize
                else:
                    w, h = (self.width, self.height)
                color = self.page.comboBox_color.currentText().lower()

                if self.filterType == 0:  # Spectrum
                    if self.amplitude == 0:
                        amplitude = "sqrt"
                    elif self.amplitude == 1:
                        amplitude = "cbrt"
                    elif self.amplitude == 2:
                        amplitude = "4thrt"
                    elif self.amplitude == 3:
                        amplitude = "5thrt"
                    elif self.amplitude == 4:
                        amplitude = "lin"
                    elif self.amplitude == 5:
                        amplitude = "log"
                    filter_ = (
                        f"showspectrum=s={w}x{h}:"
                        "slide=scroll:"
                        f"win_func={self.page.comboBox_window.currentText()}:"
                        f"color={color}:"
                        f"scale={amplitude},"
                        "colorkey=color=black:"
                        "similarity=0.1:blend=0.5"
                    )
                elif self.filterType == 1:  # Histogram
                    if self.amplitude1 == 0:
                        amplitude = "log"
                    elif self.amplitude1 == 1:
                        amplitude = "lin"
                    if self.display == 0:
                        display = "log"
                    elif self.display == 1:
                        display = "sqrt"
                    elif self.display == 2:
                        display = "cbrt"
                    elif self.display == 3:
                        display = "lin"
                    elif self.display == 4:
                        display = "rlog"
                    filter_ = (
                        f'ahistogram=r={str(self.settings.value("outputFrameRate"))}:'
                        f"s={w}x{h}:"
                        "dmode=separate:"
                        f"ascale={amplitude}:"
                        f"scale={display}"
                    )
                elif self.filterType == 2:  # Vector Scope
                    if self.amplitude2 == 0:
                        amplitude = "log"
                    elif self.amplitude2 == 1:
                        amplitude = "sqrt"
                    elif self.amplitude2 == 2:
                        amplitude = "cbrt"
                    elif self.amplitude2 == 3:
                        amplitude = "lin"
                    m = self.page.comboBox_mode.currentText()
                    filter_ = (
                        f"avectorscope=s={w}x{h}:"
                        f'draw={"line" if self.draw else "dot"}:'
                        f"m={m}:"
                        f"scale={amplitude}:"
                        f"zoom={str(self.zoom)}"
                    )
                elif self.filterType == 3:  # Musical Scale
                    filter_ = (
                        f'showcqt=r={str(self.settings.value("outputFrameRate"))}:'
                        f"s={w}x{h}:"
                        "count=30:"
                        "text=0:"
                        f"tc={str(self.tc)},"
                        "colorkey=color=black:"
                        "similarity=0.1:blend=0.5"
                    )
                elif self.filterType == 4:  # Phase
                    filter_ = (
                        f'aphasemeter=r={str(self.settings.value("outputFrameRate"))}:'
                        f"s={w}x{h}:"
                        "video=1 [atrash][vtmp1]; "
                        "[atrash] anullsink; "
                        "[vtmp1] colorkey=color=black:"
                        "similarity=0.1:blend=0.5, "
                        "crop=in_w/8:in_h:(in_w/8)*7:0  "
                    )
                return filter_

            if self.filterType < 2:
                exampleSnd = exampleSound("freq")
            elif self.filterType == 2 or self.filterType == 4:
                exampleSnd = exampleSound("stereo")
            elif self.filterType == 3:
                exampleSnd = exampleSound("white")
            compression = "compand=gain=4," if self.compress else ""
            aformat = (
                "aformat=channel_layouts=mono,"
                if self.mono and self.filterType not in (2, 4)
                else ""
            )
            filter_ = getFilterComplexCommandForType()
            hflip = "hflip, " if self.mirror else ""
            trim = (
                "trim=start=%s:end=%s, "
                % (
                    "{0:.3f}".format(startPt + 12),
                    "{0:.3f}".format(startPt + 12.5),
                )
                if preview
                else ""
            )
            scale_ = "scale=%sx%s" % scale(self.scale, self.width, self.height, str)
            hue = (
                ", hue=h=%s:s=10" % str(self.hue)
                if self.hue > 0 and self.filterType != 3
                else ""
            )
            convolution = (
                ", convolution=-2 -1 0 -1 1 1 0 1 2:-2 -1 0 -1 1 1 0 1 2:-2 -1 0 -1 1 1 0 1 2:-2 -1 0 -1 1 1 0 1 2"
                if self.filterType == 3
                else ""
            )

            return (
                f"{exampleSnd if preview and genericPreview else '[0:a] '}"
                f"{compression}{aformat}{filter_} [v1]; "
                f"[v1] {hflip}{trim}{scale_}{hue}{convolution} [v]"
            )

        return [
            "-filter_complex",
            getFilterComplexCommand(),
            "-map",
            "[v]",
        ]

    def updateChunksize(self):
        width, height = scale(self.scale, self.width, self.height, int)
        oldChunkSize = int(self.chunkSize)
        self.chunkSize = 4 * width * height
        changed = self.chunkSize != oldChunkSize
        return changed

    def finalizeFrame(self, imageData):
        try:
            image = Image.frombytes(
                "RGBA",
                scale(self.scale, self.width, self.height, int),
                imageData,
            )
            self._image = image
        except ValueError:
            image = self._image

        frame = BlankFrame(self.width, self.height)
        frame.paste(image, box=(self.x, self.y))
        return frame



from PyQt6 import QtCore
import os
from pytest import fixture
from pytestqt import qtbot
from avp.gui.mainwindow import MainWindow
from . import getTestDataPath, window, settings


def test_mainwindow_init_with_project(qtbot, settings):
    window = MainWindow(getTestDataPath("config/projects/testproject.avp"), None)
    qtbot.addWidget(window)
    assert window.core.selectedComponents[0].name == "Classic Visualizer"
    assert window.core.selectedComponents[1].name == "Color"


def test_mainwindow_clear(qtbot, window):
    """MainWindow.clear() gives us a clean slate"""
    assert len(window.core.selectedComponents) == 0


def test_mainwindow_presetDir_in_tests(qtbot, window):
    """`presetDir` is the filepath on which "Import Preset" file picker opens"""
    assert os.path.basename(window.core.settings.value("presetDir")) == "presets"


def test_mainwindow_openProject(qtbot, window):
    """Open testproject.avp using MainWindow.openProject()"""
    window.openProject(getTestDataPath("config/projects/testproject.avp"), prompt=False)
    assert len(window.core.selectedComponents) == 2


def test_mainwindow_newProject_without_unsaved_changes(qtbot, window):
    """Starting new project without unsaved changes"""
    didCallFunction = False

    def captureFunction(*args, **kwargs):
        nonlocal didCallFunction
        didCallFunction = True

    window.createNewProject(prompt=False)
    assert not didCallFunction
    assert len(window.core.selectedComponents) == 0


def test_mainwindow_newProject_with_unsaved_changes(qtbot, window):
    """Starting new project with unsaved changes"""
    didCallFunction = False

    def captureFunction(*args, **kwargs):
        nonlocal didCallFunction
        didCallFunction = True

    window.openSaveChangesDialog = captureFunction
    window.createNewProject(prompt=True)
    assert didCallFunction



from avp.command import Command
from pytestqt import qtbot
from pytest import fixture
from . import (
    imageDataSum,
    command,
    preFrameRender,
    audioData,
)


@fixture
def coreWithSpectrumComp(qtbot, command):
    """Fixture providing a Command object with Spectrum component added"""
    command.settings.setValue("outputHeight", 1080)
    command.settings.setValue("outputWidth", 1920)
    command.core.insertComponent(0, command.core.moduleIndexFor("Spectrum"), command)
    yield command.core


def test_comp_spectrum_previewRender(coreWithSpectrumComp):
    comp = coreWithSpectrumComp.selectedComponents[0]
    image = comp.previewRender()
    assert imageDataSum(image) == 71992628


def test_comp_spectrum_renderFrame(coreWithSpectrumComp, audioData):
    comp = coreWithSpectrumComp.selectedComponents[0]
    preFrameRender(audioData, comp)
    image = comp.frameRender(0)
    comp.postFrameRender()
    assert imageDataSum(image) == 117



import numpy
from PIL import Image, ImageDraw

from ..libcomponent import BaseComponent
from ..toolkit.frame import BlankFrame, FloodFrame
from ..toolkit.visualizer import createSpectrumArray


class Component(BaseComponent):
    name = "Classic Visualizer"
    version = "1.2.0"

    def names(*args):
        return ["Original"]

    def properties(self):
        props = ["pcm"]
        if self.invert:
            props.append("composite")
        return props

    def widget(self, *args):
        self.scale = 20
        self.bars = 63
        self.y = 0
        super().widget(*args)

        self.page.comboBox_visLayout.addItem("Classic")
        self.page.comboBox_visLayout.addItem("Split")
        self.page.comboBox_visLayout.addItem("Bottom")
        self.page.comboBox_visLayout.addItem("Top")
        self.page.comboBox_visLayout.setCurrentIndex(0)

        self.trackWidgets(
            {
                "visColor": self.page.lineEdit_visColor,
                "layout": self.page.comboBox_visLayout,
                "scale": self.page.spinBox_scale,
                "y": self.page.spinBox_y,
                "smooth": self.page.spinBox_sensitivity,
                "bars": self.page.spinBox_bars,
                "invert": self.page.checkBox_invert,
            },
            colorWidgets={
                "visColor": self.page.pushButton_visColor,
            },
            relativeWidgets=[
                "y",
            ],
        )

    def previewRender(self, frame=None):
        spectrum = numpy.fromfunction(
            lambda x: float(self.scale) / 2500 * (x - 128) ** 2,
            (255,),
            dtype="int16",
        )
        return self.drawBars(
            self.width,
            self.height,
            spectrum,
            self.visColor,
            self.layout,
            frame,
        )

    def preFrameRender(self, **kwargs):
        super().preFrameRender(**kwargs)
        smoothConstantDown = 0.08 if not self.smooth else self.smooth / 15
        smoothConstantUp = 0.8 if not self.smooth else self.smooth / 15
        self.spectrumArray = createSpectrumArray(
            self,
            self.completeAudioArray,
            self.sampleSize,
            smoothConstantDown,
            smoothConstantUp,
            self.scale,
            self.progressBarUpdate,
            self.progressBarSetText,
        )

    def frameRender(self, frameNo, frame=None):
        arrayNo = frameNo * self.sampleSize
        return self.drawBars(
            self.width,
            self.height,
            self.spectrumArray[arrayNo],
            self.visColor,
            self.layout,
            frame,
        )

    def drawBars(self, width, height, spectrum, color, layout, frame):
        bigYCoord = height - height / 8
        smallYCoord = height / 1200
        bigXCoord = width / (self.bars + 1)
        middleXCoord = bigXCoord / 2
        smallXCoord = bigXCoord / 4

        imTop = BlankFrame(width, height)
        draw = ImageDraw.Draw(imTop)
        r, g, b = color
        color2 = (r, g, b, 125)

        for i in range(self.bars):
            # draw outline behind rectangles if not inverted
            if frame is None:
                x0 = middleXCoord + i * bigXCoord
                y0 = bigYCoord + smallXCoord
                x1 = middleXCoord + i * bigXCoord + bigXCoord
                y1 = (
                    bigYCoord
                    + smallXCoord
                    - spectrum[i * 4] * smallYCoord
                    - middleXCoord
                )
                selection = (
                    x0,
                    y0 if y0 < y1 else y1,
                    x1 if x1 > x0 else x0,
                    y1 if y0 < y1 else y0,
                )
                draw.rectangle(
                    selection,
                    fill=color2,
                )

            x0 = middleXCoord + smallXCoord + i * bigXCoord
            y0 = bigYCoord
            x1 = middleXCoord + smallXCoord + i * bigXCoord + middleXCoord
            y1 = bigYCoord - spectrum[i * 4] * smallYCoord
            selection = (
                x0,
                y0 if y0 < y1 else y1,
                x1 if x1 > x0 else x0,
                y1 if y0 < y1 else y0,
            )
            # fill rectangle if not inverted
            draw.rectangle(
                selection,
                fill=color if frame is None else (0, 0, 0, 0),
                outline=color,
                width=int(x1 - x0),
            )

        imBottom = imTop.transpose(Image.Transpose.FLIP_TOP_BOTTOM)

        im = BlankFrame(width, height)

        if layout == 0:  # Classic
            y = self.y - int(height / 100 * 43)
            im.paste(imTop, (0, y), mask=imTop)
            y = self.y + int(height / 100 * 43)
            im.paste(imBottom, (0, y), mask=imBottom)

        if layout == 1:  # Split
            y = self.y + int(height / 100 * 10)
            im.paste(imTop, (0, y), mask=imTop)
            y = self.y - int(height / 100 * 10)
            im.paste(imBottom, (0, y), mask=imBottom)

        if layout == 2:  # Bottom
            y = self.y + int(height / 100 * 10)
            im.paste(imTop, (0, y), mask=imTop)

        if layout == 3:  # Top
            y = self.y - int(height / 100 * 10)
            im.paste(imBottom, (0, y), mask=imBottom)

        if frame is None:
            return im
        f = FloodFrame(width, height, color)
        f.paste(frame, (0, 0), mask=im)
        return f

    def command(self, arg):
        if "=" in arg:
            key, arg = arg.split("=", 1)
            try:
                if key == "color":
                    self.page.lineEdit_visColor.setText(arg)
                    return
                elif key == "layout":
                    if arg == "classic":
                        self.page.comboBox_visLayout.setCurrentIndex(0)
                    elif arg == "split":
                        self.page.comboBox_visLayout.setCurrentIndex(1)
                    elif arg == "bottom":
                        self.page.comboBox_visLayout.setCurrentIndex(2)
                    elif arg == "top":
                        self.page.comboBox_visLayout.setCurrentIndex(3)
                    return
                elif key == "scale":
                    arg = int(arg)
                    self.page.spinBox_scale.setValue(arg)
                    return
                elif key == "y":
                    arg = int(arg)
                    self.page.spinBox_y.setValue(arg)
                    return
            except ValueError:
                print("You must enter a number.")
                quit(1)
        super().command(arg)

    def commandHelp(self):
        print("Give a layout name:\n    layout=[classic/split/bottom/top]")
        print("Specify a color:\n    color=255,255,255")
        print("Visualizer scale (20 is default):\n    scale=number")
        print("Y position:\n    y=number")



import sys
from avp.cli import main
if __name__ == "__main__":
    if sys.argv[0].endswith("-script.pyw"):
        sys.argv[0] = sys.argv[0][:-11]
    elif sys.argv[0].endswith(".exe"):
        sys.argv[0] = sys.argv[0][:-4]
    sys.exit(main())



from PIL import Image
from PyQt6 import QtWidgets
import os
import subprocess
import logging

from ..libcomponent import BaseComponent
from ..toolkit.frame import BlankFrame, scale
from ..toolkit.ffmpeg import openPipe, closePipe, testAudioStream, FfmpegVideo


log = logging.getLogger("AVP.Components.Video")


class Component(BaseComponent):
    name = "Video"
    version = "1.0.0"

    def widget(self, *args):
        self.videoPath = ""
        self.badAudio = False
        self.x = 0
        self.y = 0
        self.loopVideo = False
        super().widget(*args)
        self._image = BlankFrame(self.width, self.height)
        self.page.pushButton_video.clicked.connect(self.pickVideo)
        self.trackWidgets(
            {
                "videoPath": self.page.lineEdit_video,
                "loopVideo": self.page.checkBox_loop,
                "useAudio": self.page.checkBox_useAudio,
                "distort": self.page.checkBox_distort,
                "scale": self.page.spinBox_scale,
                "volume": self.page.spinBox_volume,
                "xPosition": self.page.spinBox_x,
                "yPosition": self.page.spinBox_y,
            },
            presetNames={
                "videoPath": "video",
                "loopVideo": "loop",
                "xPosition": "x",
                "yPosition": "y",
            },
            relativeWidgets=[
                "xPosition",
                "yPosition",
            ],
        )

    def update(self):
        if self.page.checkBox_useAudio.isChecked():
            self.page.label_volume.setEnabled(True)
            self.page.spinBox_volume.setEnabled(True)
        else:
            self.page.label_volume.setEnabled(False)
            self.page.spinBox_volume.setEnabled(False)

    def previewRender(self):
        self.updateChunksize()
        frame = self.getPreviewFrame(self.width, self.height)
        if not frame:
            return BlankFrame(self.width, self.height)
        else:
            return frame

    def properties(self):
        props = []
        outputFile = None
        if hasattr(self.parent, "lineEdit_outputFile"):
            # check only happens in GUI mode
            outputFile = self.parent.lineEdit_outputFile.text()

        if not self.videoPath:
            self.lockError("There is no video selected.")
        elif not os.path.exists(self.videoPath):
            self.lockError("The video selected does not exist!")
        elif outputFile and os.path.realpath(self.videoPath) == os.path.realpath(
            outputFile
        ):
            self.lockError("Input and output paths match.")

        if self.useAudio:
            props.append("audio")
            if not testAudioStream(self.videoPath) and self.error() is None:
                self.lockError("Could not identify an audio stream in this video.")

        return props

    def audio(self):
        params = {}
        if self.volume != 1.0:
            params["volume"] = "=%s:replaygain_noclip=0" % str(self.volume)
        return (self.videoPath, params)

    def preFrameRender(self, **kwargs):
        super().preFrameRender(**kwargs)
        self.updateChunksize()
        self.video = (
            FfmpegVideo(
                inputPath=self.videoPath,
                filter_=self.makeFfmpegFilter(),
                width=self.width,
                height=self.height,
                chunkSize=self.chunkSize,
                frameRate=int(self.settings.value("outputFrameRate")),
                parent=self.parent,
                loopVideo=self.loopVideo,
                component=self,
            )
            if os.path.exists(self.videoPath)
            else None
        )

    def frameRender(self, frameNo):
        if FfmpegVideo.threadError is not None:
            raise FfmpegVideo.threadError
        return self.finalizeFrame(self.video.frame(frameNo))

    def postFrameRender(self):
        closePipe(self.video.pipe)

    def pickVideo(self):
        imgDir = self.settings.value("componentDir", os.path.expanduser("~"))
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self.page,
            "Choose Video",
            imgDir,
            "Video Files (%s)" % " ".join(self.core.videoFormats),
        )
        if filename:
            self.settings.setValue("componentDir", os.path.dirname(filename))
            self.mergeUndo = False
            self.page.lineEdit_video.setText(filename)
            self.mergeUndo = True

    def getPreviewFrame(self, width, height):
        if not self.videoPath or not os.path.exists(self.videoPath):
            return

        command = [
            self.core.FFMPEG_BIN,
            "-thread_queue_size",
            "512",
            "-i",
            self.videoPath,
            "-f",
            "image2pipe",
            "-pix_fmt",
            "rgba",
        ]
        command.extend(self.makeFfmpegFilter())
        command.extend(
            [
                "-codec:v",
                "rawvideo",
                "-",
                "-ss",
                "90",
                "-frames:v",
                "1",
            ]
        )

        if self.core.logEnabled:
            logFilename = os.path.join(
                self.core.logDir, "preview_%s.log" % str(self.compPos)
            )
            log.debug("Creating ffmpeg process (log at %s)" % logFilename)
            with open(logFilename, "w") as logf:
                logf.write(" ".join(command) + "\n\n")
            with open(logFilename, "a") as logf:
                pipe = openPipe(
                    command,
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.PIPE,
                    stderr=logf,
                    bufsize=10**8,
                )
        else:
            pipe = openPipe(
                command,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                bufsize=10**8,
            )

        byteFrame = pipe.stdout.read(self.chunkSize)
        closePipe(pipe)

        frame = self.finalizeFrame(byteFrame)
        return frame

    def makeFfmpegFilter(self):
        return [
            "-filter_complex",
            "[0:v] scale=%s:%s" % scale(self.scale, self.width, self.height, str),
        ]

    def updateChunksize(self):
        if self.scale != 100 and not self.distort:
            width, height = scale(self.scale, self.width, self.height, int)
        else:
            width, height = self.width, self.height
        self.chunkSize = 4 * width * height

    def command(self, arg):
        if "=" in arg:
            key, arg = arg.split("=", 1)
            if key == "path" and os.path.exists(arg):
                if "*%s" % os.path.splitext(arg)[1] in self.core.videoFormats:
                    self.page.lineEdit_video.setText(arg)
                    self.page.spinBox_scale.setValue(100)
                    self.page.checkBox_loop.setChecked(True)
                    return
                else:
                    print("Not a supported video format")
                    quit(1)
        elif arg == "audio":
            if not self.page.lineEdit_video.text():
                print("'audio' option must follow a video selection")
                quit(1)
            self.page.checkBox_useAudio.setChecked(True)
            return
        super().command(arg)

    def commandHelp(self):
        print("Load a video:\n    path=/filepath/to/video.mp4")
        print("Using audio:\n    path=/filepath/to/video.mp4 audio")

    def finalizeFrame(self, imageData):
        try:
            if self.distort:
                image = Image.frombytes("RGBA", (self.width, self.height), imageData)
            else:
                image = Image.frombytes(
                    "RGBA",
                    scale(self.scale, self.width, self.height, int),
                    imageData,
                )
            self._image = image
        except ValueError:
            # use last good frame
            image = self._image

        if self.scale != 100 or self.xPosition != 0 or self.yPosition != 0:
            frame = BlankFrame(self.width, self.height)
            frame.paste(image, box=(self.xPosition, self.yPosition))
        else:
            frame = image
        return frame



from PIL import Image, ImageChops
from PyQt6.QtGui import QColor
import os
import subprocess
import logging

from ..libcomponent import BaseComponent
from ..toolkit.visualizer import createSpectrumArray
from ..toolkit.frame import BlankFrame, scale
from ..toolkit.ffmpeg import (
    openPipe,
    closePipe,
    getAudioDuration,
    FfmpegVideo,
    exampleSound,
)


log = logging.getLogger("AVP.Components.Waveform")


class Component(BaseComponent):
    name = "Waveform"
    version = "2.0.0"

    @property
    def updateInterval(self):
        """How many frames from FFmpeg are ignored between each final frame"""
        return 100 - self.speed

    def properties(self):
        return [] if self.speed == 100 else ["pcm"]

    def widget(self, *args):
        super().widget(*args)
        self._image = BlankFrame(self.width, self.height)

        if hasattr(self.parent, "lineEdit_audioFile"):
            self.parent.lineEdit_audioFile.textChanged.connect(self.update)

        self.trackWidgets(
            {
                "color": self.page.lineEdit_color,
                "mode": self.page.comboBox_mode,
                "amplitude": self.page.comboBox_amplitude,
                "x": self.page.spinBox_x,
                "y": self.page.spinBox_y,
                "mirror": self.page.checkBox_mirror,
                "scale": self.page.spinBox_scale,
                "opacity": self.page.spinBox_opacity,
                "compress": self.page.checkBox_compress,
                "mono": self.page.checkBox_mono,
                "speed": self.page.spinBox_speed,
            },
            colorWidgets={
                "color": self.page.pushButton_color,
            },
            relativeWidgets=[
                "x",
                "y",
            ],
        )

    def previewRender(self):
        self.updateChunksize()
        frame = self.getPreviewFrame(self.width, self.height)
        if not frame:
            return BlankFrame(self.width, self.height)
        else:
            return frame

    def preFrameRender(self, **kwargs):
        self._fadingImage = None
        self._prevImage = None
        self._currImage = None
        self._lastUpdatedFrame = 0
        super().preFrameRender(**kwargs)
        self.updateChunksize()
        w, h = scale(self.scale, self.width, self.height, str)
        self.video = FfmpegVideo(
            inputPath=self.audioFile,
            filter_=self.makeFfmpegFilter(),
            width=w,
            height=h,
            chunkSize=self.chunkSize,
            frameRate=int(self.settings.value("outputFrameRate")),
            parent=self.parent,
            component=self,
            debug=True,
        )
        if self.speed == 100:
            return
        self.spectrumArray = createSpectrumArray(
            self,
            self.completeAudioArray,
            self.sampleSize,
            0.08,
            0.8,
            20,
            self.progressBarUpdate,
            self.progressBarSetText,
        )

    def frameRender(self, frameNo):
        if FfmpegVideo.threadError is not None:
            raise FfmpegVideo.threadError
        newFrame = self.finalizeFrame(self.video.frame(frameNo))
        if self.speed == 100:
            return newFrame
        frameDiff = 0 if frameNo == 0 else frameNo % self.updateInterval
        peaks = [
            self.spectrumArray[frameNo * self.sampleSize][i * 4] for i in range(64)
        ]
        peakValue = 70 - (max(*peaks) - min(*peaks))
        isValidPeak = (
            peakValue > 27
            and frameNo - self._lastUpdatedFrame > self.updateInterval / 2
        )
        if frameDiff == 0 or isValidPeak:
            self._lastUpdatedFrame = frameNo
            self._fadingImage = self._prevImage
            self._prevImage = self._image
            self._currImage = newFrame
        usualAlpha = 0.0 + (1 / self.updateInterval) * frameDiff
        alpha = max(
            0.1
            + (
                1
                / max(
                    10,
                    peakValue,
                )
            ),
            usualAlpha,
        )
        baseImage = self._prevImage
        if self._fadingImage is not None:
            # fade away the old previous frame from ages ago
            baseImage = ImageChops.blend(
                self._prevImage, self._fadingImage, max(0.0, 0.9 - usualAlpha)
            )
        blendedImage = ImageChops.blend(
            baseImage,
            ImageChops.lighter(self._prevImage, newFrame),
            alpha,
        )
        baseImage.paste(blendedImage, (0, 0), mask=blendedImage)
        return Image.alpha_composite(self._currImage, baseImage)

    def postFrameRender(self):
        closePipe(self.video.pipe)

    def getPreviewFrame(self, width, height):
        genericPreview = self.settings.value("pref_genericPreview")
        startPt = 0
        if not genericPreview:
            inputFile = self.parent.lineEdit_audioFile.text()
            if not inputFile or not os.path.exists(inputFile):
                return
            duration = getAudioDuration(inputFile)
            if not duration:
                return
            startPt = duration / 3
            if startPt + 3 > duration:
                startPt += startPt - 3

        command = [
            self.core.FFMPEG_BIN,
            "-thread_queue_size",
            "512",
            "-r",
            str(self.settings.value("outputFrameRate")),
            "-ss",
            "{0:.3f}".format(startPt),
            "-i",
            self.core.junkStream if genericPreview else inputFile,
            "-f",
            "image2pipe",
            "-pix_fmt",
            "rgba",
        ]
        command.extend(self.makeFfmpegFilter(preview=True, startPt=startPt))
        command.extend(
            [
                "-an",
                "-s:v",
                "%sx%s" % scale(self.scale, self.width, self.height, str),
                "-codec:v",
                "rawvideo",
                "-",
                "-frames:v",
                "1",
            ]
        )
        if self.core.logEnabled:
            logFilename = os.path.join(
                self.core.logDir, "preview_%s.log" % str(self.compPos)
            )
            log.debug("Creating ffmpeg log at %s", logFilename)
            with open(logFilename, "w") as logf:
                logf.write(" ".join(command) + "\n\n")
            with open(logFilename, "a") as logf:
                pipe = openPipe(
                    command,
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.PIPE,
                    stderr=logf,
                    bufsize=10**8,
                )
        else:
            pipe = openPipe(
                command,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                bufsize=10**8,
            )
        byteFrame = pipe.stdout.read(self.chunkSize)
        closePipe(pipe)

        frame = self.finalizeFrame(byteFrame)
        return frame

    def makeFfmpegFilter(self, preview=False, startPt=0):
        w, h = scale(self.scale, self.width, self.height, str)
        if self.amplitude == 0:
            amplitude = "log"
        elif self.amplitude == 1:
            amplitude = "sqrt"
        elif self.amplitude == 2:
            amplitude = "cbrt"
        elif self.amplitude == 3:
            amplitude = "lin"
        hexcolor = QColor(*self.color).name()
        opacity = "{0:.1f}".format(self.opacity / 100)
        genericPreview = self.settings.value("pref_genericPreview")
        if self.mode > 1:
            filter_ = (
                "showwaves="
                f'r={str(self.settings.value("outputFrameRate"))}:'
                f's={self.settings.value("outputWidth")}x{self.settings.value("outputHeight")}:'
                f'mode={self.page.comboBox_mode.currentText().lower() if self.mode != 3 else "p2p"}:'
                f"colors={hexcolor}@{opacity}:scale={amplitude}"
            )
        elif self.mode < 2:
            filter_ = (
                f'showfreqs=s={str(self.settings.value("outputWidth"))}x{str(self.settings.value("outputHeight"))}:'
                f'mode={"line" if self.mode == 0 else "bar"}:'
                f"colors={hexcolor}@{opacity}"
                f":ascale={amplitude}:fscale={'log' if self.mono else 'lin'}"
            )

        baselineHeight = int(self.height * (4 / 1080))
        return [
            "-filter_complex",
            f"{exampleSound('wave', extra='') if preview and genericPreview else '[0:a] '}"
            f"{'compand=gain=4,' if self.compress else ''}"
            f"{'aformat=channel_layouts=mono,' if self.mono and self.mode < 3 else ''}"
            f"{filter_}"
            f"{', drawbox=x=(iw-w)/2:y=(ih-h)/2:w=iw:h=%s:color=%s@%s' % (baselineHeight, hexcolor, opacity) if self.mode < 2 else ''}"
            f"{', hflip' if self.mirror else''}"
            " [v1]; "
            "[v1] scale=%s:%s%s [v]"
            % (
                w,
                h,
                ", trim=duration=%s" % "{0:.3f}".format(startPt + 3) if preview else "",
            ),
            "-map",
            "[v]",
        ]

    def updateChunksize(self):
        width, height = scale(self.scale, self.width, self.height, int)
        self.chunkSize = 4 * width * height

    def finalizeFrame(self, imageData):
        try:
            image = Image.frombytes(
                "RGBA",
                scale(self.scale, self.width, self.height, int),
                imageData,
            )
            self._image = image
        except ValueError:
            image = self._image
        if self.scale != 100 or self.x != 0 or self.y != 0:
            frame = BlankFrame(self.width, self.height)
            frame.paste(image, box=(self.x, self.y))
        else:
            frame = image
        return frame



"""
Thread that runs to create QImages for MainWindow's preview label.
Processes a queue of component lists.
"""

from PyQt6 import QtCore, QtGui, uic
from PyQt6.QtCore import pyqtSignal, pyqtSlot
from PIL import Image
from PIL.ImageQt import ImageQt
from queue import Queue, Empty
import os
import logging

from ..toolkit.frame import Checkerboard
from ..toolkit import disableWhenOpeningProject


log = logging.getLogger("AVP.Gui.PreviewThread")


class Worker(QtCore.QObject):

    imageCreated = pyqtSignal(QtGui.QImage)
    error = pyqtSignal(str)

    def __init__(self, core, settings, queue):
        super().__init__()
        self.core = core
        self.settings = settings
        width = int(self.settings.value("outputWidth"))
        height = int(self.settings.value("outputHeight"))
        self.queue = queue
        self.background = Checkerboard(width, height)

    @disableWhenOpeningProject
    @pyqtSlot(list)
    def createPreviewImage(self, components):
        dic = {
            "components": components,
        }
        self.queue.put(dic)
        log.debug("Preview thread id: {}".format(int(QtCore.QThread.currentThreadId())))

    @pyqtSlot()
    def process(self):
        try:
            nextPreviewInformation = self.queue.get(block=False)
            while self.queue.qsize() >= 2:
                try:
                    self.queue.get(block=False)
                except Empty:
                    continue
            width = int(self.settings.value("outputWidth"))
            height = int(self.settings.value("outputHeight"))
            if self.background.width != width or self.background.height != height:
                self.background = Checkerboard(width, height)

            frame = self.background.copy()
            log.info("Creating new preview frame")
            components = nextPreviewInformation["components"]
            for component in reversed(components):
                try:
                    component.lockSize(width, height)
                    if "composite" in component.properties():
                        newFrame = component.previewRender(frame)
                    else:
                        newFrame = component.previewRender()
                    component.unlockSize()
                    frame = Image.alpha_composite(frame, newFrame)

                except (AttributeError, ValueError) as e:
                    errMsg = (
                        "Bad frame returned by %s's preview renderer. "
                        "%s. New frame %s."
                        % (
                            str(component),
                            str(e).capitalize(),
                            (
                                "is None"
                                if newFrame is None
                                else "size was %s*%s; should be %s*%s"
                                % (newFrame.width, newFrame.height, width, height)
                            ),
                        )
                    )
                    log.critical(errMsg)
                    self.error.emit(errMsg)
                    break
                except RuntimeError as e:
                    log.error(str(e))
            else:
                # We must store a reference to this QImage
                # or else Qt will garbage-collect it on the C++ side
                self.frame = ImageQt(frame)
                self.imageCreated.emit(QtGui.QImage(self.frame))

        except Empty:
            True



from avp.toolkit.visualizer import transformData
from pytestqt import qtbot
from pytest import fixture, mark
from . import audioData, command, imageDataSum, preFrameRender


sampleSize = 1470  # 44100 / 30 = 1470


def createSpectrumArray(audioData):
    """Creates enough `spectrumArray` for one call to Component.drawBars()"""
    spectrumArray = {0: transformData(0, audioData[0], sampleSize, 0.08, 0.8, None, 20)}
    for i in range(sampleSize, len(audioData[0]), sampleSize):
        spectrumArray[i] = transformData(
            i,
            audioData[0],
            sampleSize,
            0.08,
            0.8,
            spectrumArray[i - sampleSize].copy(),
            20,
        )
    return spectrumArray


@fixture
def coreWithClassicComp(qtbot, command):
    """Fixture providing a Command object with Classic Visualizer component added"""
    command.core.insertComponent(
        0, command.core.moduleIndexFor("Classic Visualizer"), command
    )
    yield command.core


def test_comp_classic_added(coreWithClassicComp):
    """Add Classic Visualizer to core"""
    assert len(coreWithClassicComp.selectedComponents) == 1


def test_comp_classic_removed(coreWithClassicComp):
    """Remove Classic Visualizer from core"""
    coreWithClassicComp.removeComponent(0)
    assert len(coreWithClassicComp.selectedComponents) == 0


@mark.parametrize("layout", (0, 1, 2, 3))
def test_comp_classic_drawBars(coreWithClassicComp, audioData, layout):
    """Call drawBars after creating audio spectrum data manually."""
    spectrumArray = createSpectrumArray(audioData)
    comp = coreWithClassicComp.selectedComponents[0]
    image = comp.drawBars(
        1920, 1080, spectrumArray[sampleSize * 4], (0, 0, 0), layout, None
    )
    imageSize = 37872316
    assert imageDataSum(image) == imageSize if layout < 2 else imageSize / 2


def test_comp_classic_drawBars_using_preFrameRender(coreWithClassicComp, audioData):
    """Call drawBars after creating audio spectrum data using preFrameRender."""
    comp = coreWithClassicComp.selectedComponents[0]
    preFrameRender(audioData, comp)
    image = comp.drawBars(
        1920,
        1080,
        coreWithClassicComp.selectedComponents[0].spectrumArray[sampleSize * 4],
        (0, 0, 0),
        0,
        None,
    )
    assert imageDataSum(image) == 37872316


def test_comp_classic_command_layout(coreWithClassicComp):
    comp = coreWithClassicComp.selectedComponents[0]
    comp.command("layout=top")
    assert comp.layout == 3


def test_comp_classic_command_color(coreWithClassicComp):
    comp = coreWithClassicComp.selectedComponents[0]
    comp.command("color=111,111,111")
    assert comp.visColor == (111, 111, 111)


def test_comp_classic_command_preset(coreWithClassicComp):
    comp = coreWithClassicComp.selectedComponents[0]
    saveValueStore = comp.savePreset()
    saveValueStore["preset"] = "testPreset"
    coreWithClassicComp.createPresetFile(
        comp.name, comp.version, "testPreset", saveValueStore
    )
    comp.command("preset=testPreset")
    assert comp.currentPreset == "testPreset"


def test_comp_classic_loadPreset(coreWithClassicComp):
    comp = coreWithClassicComp.selectedComponents[0]
    comp.scale = 99
    saveValueStore = comp.savePreset()
    saveValueStore["preset"] = "testPreset"
    comp.scale = 20
    comp.loadPreset(saveValueStore, "testPreset")
    assert comp.scale == 99



from avp import __version__
from tomllib import load
import os


def test_version_number_matches():
    with open(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "..", "pyproject.toml"
        ),
        "rb",
    ) as fp:
        config = load(fp)
    assert config["project"]["version"] == __version__



from avp.command import Command
from pytestqt import qtbot
from pytest import fixture
from . import imageDataSum, command


@fixture
def coreWithColorComp(qtbot, command):
    """Fixture providing a Command object with Color component added"""
    command.settings.setValue("outputHeight", 1080)
    command.settings.setValue("outputWidth", 1920)
    command.core.insertComponent(0, command.core.moduleIndexFor("Color"), command)
    yield command.core


def test_comp_color_set_color(coreWithColorComp):
    """Set imagePath of Image component"""
    comp = coreWithColorComp.selectedComponents[0]
    comp.page.lineEdit_color1.setText("111,111,111")
    image = comp.previewRender()
    assert imageDataSum(image) == 1219276800


def test_comp_color_gradient(coreWithColorComp):
    """Test changing fill type to a gradient"""
    comp = coreWithColorComp.selectedComponents[0]
    comp.page.comboBox_fill.setCurrentIndex(1)
    comp.page.lineEdit_color1.setText("0,0,0")
    comp.page.lineEdit_color2.setText("255,255,255")
    image = comp.previewRender()
    assert imageDataSum(image) == 1849285965



from PIL import Image, ImageOps, ImageEnhance
from PyQt6 import QtWidgets
import os

from ..libcomponent import BaseComponent
from ..toolkit.frame import BlankFrame, addShadow
from ..toolkit.visualizer import createSpectrumArray


class Component(BaseComponent):
    name = "Image"
    version = "2.1.0"

    def widget(self, *args):
        super().widget(*args)

        # cache a modified image object in case we are rendering beyond frame 1
        self.existingImage = None

        self.page.pushButton_image.clicked.connect(self.pickImage)
        self.page.comboBox_resizeMode.addItem("Scale")
        self.page.comboBox_resizeMode.addItem("Cover")
        self.page.comboBox_resizeMode.addItem("Stretch")
        self.page.comboBox_resizeMode.setCurrentIndex(0)
        self.trackWidgets(
            {
                "imagePath": self.page.lineEdit_image,
                "scale": self.page.spinBox_scale,
                "rotate": self.page.spinBox_rotate,
                "color": self.page.spinBox_color,
                "xPosition": self.page.spinBox_x,
                "yPosition": self.page.spinBox_y,
                "resizeMode": self.page.comboBox_resizeMode,
                "mirror": self.page.checkBox_mirror,
                "respondToAudio": self.page.checkBox_respondToAudio,
                "sensitivity": self.page.spinBox_sensitivity,
                "shadow": self.page.checkBox_shadow,
            },
            presetNames={
                "imagePath": "image",
                "xPosition": "x",
                "yPosition": "y",
            },
            relativeWidgets=["xPosition", "yPosition", "scale"],
        )

    def update(self):
        self.page.spinBox_sensitivity.setEnabled(
            self.page.checkBox_respondToAudio.isChecked()
        )
        self.page.spinBox_scale.setEnabled(
            self.page.comboBox_resizeMode.currentIndex() == 0
        )

    def previewRender(self):
        return self.drawFrame(self.width, self.height, None)

    def properties(self):
        props = ["pcm" if self.respondToAudio else "static"]
        if not os.path.exists(self.imagePath):
            props.append("error")
        return props

    def error(self):
        if not self.imagePath:
            return "There is no image selected."
        if not os.path.exists(self.imagePath):
            return "The image selected does not exist!"

    def preFrameRender(self, **kwargs):
        super().preFrameRender(**kwargs)
        if not self.respondToAudio:
            return

        # Trigger creation of new base image
        self.existingImage = None

        self.spectrumArray = createSpectrumArray(
            self,
            self.completeAudioArray,
            self.sampleSize,
            0.08,
            0.8,
            self.sensitivity,
            self.progressBarUpdate,
            self.progressBarSetText,
        )

    def frameRender(self, frameNo):
        return self.drawFrame(
            self.width,
            self.height,
            (
                None
                if not self.respondToAudio
                else self.spectrumArray[frameNo * self.sampleSize]
            ),
        )

    def drawFrame(self, width, height, dynamicScale):
        frame = BlankFrame(width, height)
        if self.imagePath and os.path.exists(self.imagePath):
            if dynamicScale is not None and self.existingImage:
                image = self.existingImage
            else:
                image = Image.open(self.imagePath)
                # Modify static image appearance
                if self.color != 100:
                    image = ImageEnhance.Color(image).enhance(float(self.color / 100))
                if self.mirror:
                    image = image.transpose(Image.Transpose.FLIP_LEFT_RIGHT)
                if self.resizeMode == 1:  # Cover
                    image = ImageOps.fit(
                        image, (width, height), Image.Resampling.LANCZOS
                    )
                elif self.resizeMode == 2:  # Stretch
                    image = image.resize((width, height), Image.Resampling.LANCZOS)
                elif self.scale != 100:  # Scale
                    newHeight = int((image.height / 100) * self.scale)
                    newWidth = int((image.width / 100) * self.scale)
                    image = image.resize(
                        (newWidth, newHeight), Image.Resampling.LANCZOS
                    )
                self.existingImage = image

            # Respond to audio
            resolutionFactor = height / 1080
            shadX = int(resolutionFactor * 1)
            shadY = int(resolutionFactor * -1)
            shadBlur = resolutionFactor * 3.50
            scale = 0
            if dynamicScale is not None:
                scale = dynamicScale[36 * 4] / 4
                shadX += int((scale / 4) * resolutionFactor)
                shadY += int((scale / 2) * resolutionFactor)
                shadBlur += (scale / 8) * resolutionFactor
                image = ImageOps.contain(
                    image,
                    (
                        image.width + int(scale / 2),
                        image.height + int(scale / 2),
                    ),
                    Image.Resampling.LANCZOS,
                )

            # Paste image at correct position
            frame.paste(
                image,
                box=(
                    self.xPosition - (0 if not self.respondToAudio else int(scale / 2)),
                    self.yPosition - (0 if not self.respondToAudio else int(scale / 2)),
                ),
            )
            if self.rotate != 0:
                frame = frame.rotate(self.rotate)
            if self.shadow:
                frame = addShadow(frame, shadBlur, shadX, shadY)

        return frame

    def postFrameRender(self):
        self.existingImage = None

    def pickImage(self):
        imgDir = self.settings.value("componentDir", os.path.expanduser("~"))
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self.page,
            "Choose Image",
            imgDir,
            "Image Files (%s)" % " ".join(self.core.imageFormats),
        )
        if filename:
            self.settings.setValue("componentDir", os.path.dirname(filename))
            self.mergeUndo = False
            self.page.lineEdit_image.setText(filename)
            self.mergeUndo = True

    def command(self, arg):
        def fail():
            print("Not a supported image format")
            quit(1)

        if "=" in arg:
            key, arg = arg.split("=", 1)
            if key == "path" and os.path.exists(arg):
                if f"*{os.path.splitext(arg)[1]}" not in self.core.imageFormats:
                    fail()
                try:
                    Image.open(arg)
                    self.page.lineEdit_image.setText(arg)
                    self.page.comboBox_resizeMode.setCurrentIndex(2)
                    return
                except OSError as e:
                    fail()
        super().command(arg)

    def commandHelp(self):
        print("Load an image:\n    path=/filepath/to/image.png")



"""
Home to the Core class which tracks program state. Used by GUI & commandline
to create a list of components and create a video thread to export.
"""

from PyQt6 import QtCore, QtGui, uic
import sys
import os
import json
from importlib import import_module
import logging

from . import toolkit

appName = "Audio Visualizer Python"
log = logging.getLogger("AVP.Core")


class Core:
    """
    MainWindow and Command module both use an instance of this class
    to store the core program state. This object tracks the components,
    talks to the components, handles opening/creating project files
    and presets, and creates the video thread to export.
    This class also stores constants as class variables.
    """

    stdoutLogLvl = logging.WARNING

    def __init__(self):
        self.importComponents()
        self.selectedComponents = []
        self.savedPresets = {}  # copies of presets to detect modification
        self.openingProject = False

    def __repr__(self):
        return "\n=~=~=~=\n".join([repr(comp) for comp in self.selectedComponents])

    def importComponents(self):
        def findComponents():
            for f in os.listdir(Core.componentsPath):
                name, ext = os.path.splitext(f)
                if name.startswith("__"):
                    continue
                elif ext == ".py":
                    yield name

        log.debug("Importing component modules")
        self.modules = [
            import_module(".components.%s" % name, __package__)
            for name in findComponents()
        ]
        # store canonical module names and indexes
        self.moduleIndexes = [i for i in range(len(self.modules))]
        self.compNames = [mod.Component.name for mod in self.modules]
        # alphabetize modules by Component name
        sortedModules = sorted(zip(self.compNames, self.modules))
        self.compNames = [y[0] for y in sortedModules]
        self.modules = [y[1] for y in sortedModules]

        # store alternative names for modules
        self.altCompNames = []
        for i, mod in enumerate(self.modules):
            if hasattr(mod.Component, "names"):
                for name in mod.Component.names():
                    self.altCompNames.append((name, i))

    def componentListChanged(self):
        for i, component in enumerate(self.selectedComponents):
            component.compPos = i

    def insertComponent(self, compPos, component, loader):
        """
        Creates a new component using these args:
        (compPos, component obj or moduleIndex, MWindow/Command obj)
        """
        if compPos < 0 or compPos > len(self.selectedComponents):
            compPos = len(self.selectedComponents)
        if len(self.selectedComponents) > 50:
            return -1
        if component is None:
            log.warning("Tried to insert non-existent component")
            return -1
        elif type(component) is int:
            # create component using module index in self.modules
            moduleIndex = int(component)
            log.debug("Creating new component from module #%s", str(moduleIndex))
            component = self.modules[moduleIndex].Component(moduleIndex, compPos, self)
            component.widget(loader)
        else:
            moduleIndex = -1
            log.debug("Inserting previously-created %s component", component.name)

        component._error.connect(loader.videoThreadError)
        self.selectedComponents.insert(compPos, component)
        if hasattr(loader, "insertComponent"):
            loader.insertComponent(compPos)

        self.componentListChanged()
        self.updateComponent(compPos)
        return compPos

    def moveComponent(self, startI, endI):
        comp = self.selectedComponents.pop(startI)
        self.selectedComponents.insert(endI, comp)

        self.componentListChanged()
        return endI

    def removeComponent(self, i):
        self.selectedComponents.pop(i)
        self.componentListChanged()

    def clearComponents(self):
        self.selectedComponents = list()
        self.componentListChanged()

    def updateComponent(self, i):
        log.debug("Auto-updating %s #%s", self.selectedComponents[i], str(i))
        self.selectedComponents[i].update(auto=True)

    def moduleIndexFor(self, compName):
        try:
            index = self.compNames.index(compName)
            return self.moduleIndexes[index]
        except ValueError:
            for altName, modI in self.altCompNames:
                if altName == compName:
                    return self.moduleIndexes[modI]

    def clearPreset(self, compIndex):
        self.selectedComponents[compIndex].currentPreset = None

    def openPreset(self, filepath, compIndex, presetName):
        """Applies a preset to a specific component"""
        saveValueStore = self.getPreset(filepath)
        if not saveValueStore:
            return False
        comp = self.selectedComponents[compIndex]
        comp.loadPreset(saveValueStore, presetName)

        self.savedPresets[presetName] = dict(saveValueStore)
        return True

    def getPreset(self, filepath):
        """Returns the preset dict stored at this filepath"""
        if not os.path.exists(filepath):
            return False
        with open(filepath, "r") as f:
            for line in f:
                saveValueStore = toolkit.presetFromString(line.strip())
                break
        return saveValueStore

    def getPresetDir(self, comp):
        """Get the preset subdir for a particular version of a component"""
        return os.path.join(Core.presetDir, comp.name, str(comp.version))

    def openProject(self, loader, filepath):
        """loader is the object calling this method which must have
        its own showMessage(**kwargs) method for displaying errors.
        """
        if not os.path.exists(filepath):
            loader.showMessage(msg="Project file not found.")
            return

        errcode, data = self.parseAvFile(filepath)
        if errcode == 0:
            self.openingProject = True
            try:
                if hasattr(loader, "window"):
                    for widget, value in data["WindowFields"]:
                        widget = eval("loader.%s" % widget)
                        with toolkit.blockSignals(widget):
                            toolkit.setWidgetValue(widget, value)

                for key, value in data["Settings"]:
                    Core.settings.setValue(key, value)
                for tup in data["Components"]:
                    name, vers, preset = tup
                    clearThis = False
                    modified = False

                    # add loaded named presets to savedPresets dict
                    if "preset" in preset and preset["preset"] is not None:
                        nam = preset["preset"]
                        filepath2 = os.path.join(Core.presetDir, name, str(vers), nam)
                        origSaveValueStore = self.getPreset(filepath2)
                        if origSaveValueStore:
                            self.savedPresets[nam] = dict(origSaveValueStore)
                            modified = not origSaveValueStore == preset
                        else:
                            # saved preset was renamed or deleted
                            clearThis = True

                    # create the actual component object & get its index
                    i = self.insertComponent(-1, self.moduleIndexFor(name), loader)
                    if i is None:
                        loader.showMessage(
                            msg=f"Component '{name}' didn't initialize correctly and had to be removed."
                        )
                        continue
                    if i == -1:
                        loader.showMessage(msg="Invalid components!")
                        break

                    try:
                        if "preset" in preset and preset["preset"] is not None:
                            self.selectedComponents[i].loadPreset(preset)
                        else:
                            self.selectedComponents[i].loadPreset(
                                preset, preset["preset"]
                            )
                    except KeyError as e:
                        log.warning(
                            "%s missing value: %s" % (self.selectedComponents[i], e)
                        )

                    if clearThis:
                        self.clearPreset(i)
                    if hasattr(loader, "updateComponentTitle"):
                        loader.updateComponentTitle(i, modified)
                self.openingProject = False
                return True
            except Exception:
                errcode = 1
                data = sys.exc_info()

        if errcode == 1:
            typ, value, tb = data
            if typ.__name__ == "KeyError":
                # probably just an old version, still loadable
                log.warning("Project file missing value: %s" % value)
                return
            if hasattr(loader, "createNewProject"):
                loader.createNewProject(prompt=False)
            msg = "%s: %s\n\n" % (typ.__name__, value)
            msg += toolkit.formatTraceback(tb)
            loader.showMessage(
                msg="Project file '%s' is corrupted." % filepath,
                showCancel=False,
                icon="Warning",
                detail=msg,
            )
            self.openingProject = False
            return False

    def parseAvFile(self, filepath):
        """
        Parses an avp (project) or avl (preset package) file.
        Returns dictionary with section names as the keys, each one
        contains a list of tuples: (compName, version, compPresetDict)
        """
        log.debug("Parsing av file: %s", filepath)
        validSections = ("Components", "Settings", "WindowFields")
        data = {sect: [] for sect in validSections}
        try:
            with open(filepath, "r") as f:

                def parseLine(line):
                    """Decides if a file line is a section header"""
                    line = line.strip()
                    newSection = ""

                    if (
                        line.startswith("[")
                        and line.endswith("]")
                        and line[1:-1] in validSections
                    ):
                        newSection = line[1:-1]

                    return line, newSection

                section = ""
                i = 0
                for line in f:
                    line, newSection = parseLine(line)
                    if newSection:
                        section = str(newSection)
                        continue
                    if line and section == "Components":
                        if i == 0:
                            lastCompName = str(line)
                            i += 1
                        elif i == 1:
                            lastCompVers = str(line)
                            i += 1
                        elif i == 2:
                            lastCompPreset = toolkit.presetFromString(line)
                            data[section].append(
                                (lastCompName, lastCompVers, lastCompPreset)
                            )
                            i = 0
                    elif line and section:
                        key, value = line.split("=", 1)
                        data[section].append((key, value.strip()))

            return 0, data
        except Exception:
            return 1, sys.exc_info()

    def importPreset(self, filepath):
        errcode, data = self.parseAvFile(filepath)
        returnList = []
        if errcode == 0:
            name, vers, preset = data["Components"][0]
            presetName = (
                preset["preset"]
                if preset["preset"]
                else os.path.basename(filepath)[:-4]
            )
            newPath = os.path.join(Core.presetDir, name, vers, presetName)
            if os.path.exists(newPath):
                return False, newPath
            preset["preset"] = presetName
            self.createPresetFile(name, vers, presetName, preset)
            return True, presetName
        elif errcode == 1:
            # TODO: an error message
            return False, ""

    def exportPreset(self, exportPath, compName, vers, origName):
        internalPath = os.path.join(Core.presetDir, compName, str(vers), origName)
        if not os.path.exists(internalPath):
            return
        if os.path.exists(exportPath):
            os.remove(exportPath)
        with open(internalPath, "r") as f:
            internalData = [line for line in f]
        try:
            saveValueStore = toolkit.presetFromString(internalData[0].strip())
            self.createPresetFile(compName, vers, origName, saveValueStore, exportPath)
            return True
        except Exception:
            return False

    def createPresetFile(self, compName, vers, presetName, saveValueStore, filepath=""):
        """Create a preset file (.avl) at filepath using args.
        Or if filepath is empty, create an internal preset using args"""
        if not filepath:
            dirname = os.path.join(Core.presetDir, compName, str(vers))
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            filepath = os.path.join(dirname, presetName)
            internal = True
        else:
            if not filepath.endswith(".avl"):
                filepath += ".avl"
            internal = False

        with open(filepath, "w") as f:
            if not internal:
                f.write("[Components]\n")
                f.write("%s\n" % compName)
                f.write("%s\n" % str(vers))
            f.write(toolkit.presetToString(saveValueStore))

    def createProjectFile(self, filepath, window=None):
        """Create a project file (.avp) using the current program state"""
        log.info("Creating %s", filepath)
        settingsKeys = [
            "componentDir",
            "inputDir",
            "outputDir",
            "presetDir",
            "projectDir",
        ]
        try:
            if not filepath.endswith(".avp"):
                filepath += ".avp"
            if os.path.exists(filepath):
                os.remove(filepath)

            with open(filepath, "w") as f:
                f.write("[Components]\n")
                for comp in self.selectedComponents:
                    saveValueStore = comp.savePreset()
                    saveValueStore["preset"] = comp.currentPreset
                    f.write("%s\n" % str(comp))
                    f.write("%s\n" % str(comp.version))
                    f.write("%s\n" % toolkit.presetToString(saveValueStore))

                f.write("\n[Settings]\n")
                for key in Core.settings.allKeys():
                    if key in settingsKeys:
                        f.write("%s=%s\n" % (key, Core.settings.value(key)))

                if window:
                    f.write("\n[WindowFields]\n")
                    f.write(
                        "lineEdit_audioFile=%s\n"
                        "lineEdit_outputFile=%s\n"
                        % (
                            window.lineEdit_audioFile.text(),
                            window.lineEdit_outputFile.text(),
                        )
                    )
            return True
        except Exception:
            return False

    def newVideoWorker(self, loader, audioFile, outputPath):
        """loader is MainWindow or Command object which must own the thread"""
        from . import video_thread

        self.videoThread = QtCore.QThread(loader)
        videoWorker = video_thread.Worker(
            loader, audioFile, outputPath, self.selectedComponents
        )
        videoWorker.moveToThread(self.videoThread)
        videoWorker.videoCreated.connect(self.stopVideoThread)

        self.videoThread.start()
        return videoWorker

    def stopVideoThread(self):
        self.videoThread.quit()
        self.videoThread.wait()

    def cancel(self):
        Core.canceled = True

    def reset(self):
        Core.canceled = False

    @classmethod
    def storeSettings(cls, dataDir=None):
        """Store settings/paths to directories as class variables"""
        from .__init__ import wd
        from .toolkit.ffmpeg import findFfmpeg

        cls.wd = wd
        dataDir = cls.getConfigPath(dataDir)
        # Windows: C:/Users/<USER>/AppData/Local/audio-visualizer
        # macOS: ~/Library/Preferences/audio-visualizer
        # Linux: ~/.config/audio-visualizer
        with open(os.path.join(wd, "encoder-options.json")) as json_file:
            encoderOptions = json.load(json_file)

        # Locate FFmpeg
        ffmpegBin = findFfmpeg()
        if not ffmpegBin:
            print("Could not find FFmpeg")

        settings = {
            "canceled": False,
            "FFMPEG_BIN": ffmpegBin,
            "dataDir": dataDir,
            "settings": QtCore.QSettings(
                os.path.join(dataDir, "settings.ini"),
                QtCore.QSettings.Format.IniFormat,
            ),
            "presetDir": os.path.join(dataDir, "presets"),
            "componentsPath": os.path.join(wd, "components"),
            "junkStream": os.path.join(wd, "gui", "background.png"),
            "encoderOptions": encoderOptions,
            "resolutions": [
                "1920x1080",
                "1280x720",
                "854x480",
            ],
            "logDir": os.path.join(dataDir, "log"),
            "logEnabled": False,
            "previewEnabled": True,
        }

        settings["videoFormats"] = toolkit.appendUppercase(
            [
                "*.mp4",
                "*.mov",
                "*.mkv",
                "*.avi",
                "*.webm",
                "*.flv",
            ]
        )
        settings["audioFormats"] = toolkit.appendUppercase(
            [
                "*.mp3",
                "*.wav",
                "*.ogg",
                "*.fla",
                "*.flac",
                "*.aac",
            ]
        )
        settings["imageFormats"] = toolkit.appendUppercase(
            [
                "*.png",
                "*.jpg",
                "*.tif",
                "*.tiff",
                "*.gif",
                "*.bmp",
                "*.ico",
                "*.xbm",
                "*.xpm",
            ]
        )

        # Register all settings as class variables
        for classvar, val in settings.items():
            setattr(cls, classvar, val)

        cls.loadDefaultSettings()
        if not os.path.exists(cls.dataDir):
            os.makedirs(cls.dataDir)
        for neededDirectory in (
            cls.presetDir,
            cls.logDir,
            cls.settings.value("projectDir"),
        ):
            if not os.path.exists(neededDirectory):
                os.mkdir(neededDirectory)
        cls.makeLogger(deleteOldLogs=True)

    @classmethod
    def loadDefaultSettings(cls):
        # settings that get saved into the ini file
        cls.defaultSettings = {
            "outputWidth": 1280,
            "outputHeight": 720,
            "outputFrameRate": 30,
            "outputAudioCodec": "AAC",
            "outputAudioBitrate": "192",
            "outputVideoCodec": "H264",
            "outputVideoBitrate": "2500",
            "outputVideoFormat": "yuv420p",
            "outputPreset": "medium",
            "outputFormat": "mp4",
            "outputContainer": "MP4",
            "projectDir": os.path.join(cls.dataDir, "projects"),
            "pref_insertCompAtTop": True,
            "pref_genericPreview": True,
            "pref_undoLimit": 10,
        }

        for parm, value in cls.defaultSettings.items():
            if cls.settings.value(parm) is None:
                cls.settings.setValue(parm, value)

        # Allow manual editing of prefs. (Surprisingly necessary as Qt seems to
        # store True as 'true' but interprets a manually-added 'true' as str.)
        for key in cls.settings.allKeys():
            if not key.startswith("pref_"):
                continue
            val = cls.settings.value(key)
            try:
                val = int(val)
            except ValueError:
                if val == "true":
                    val = True
                elif val == "false":
                    val = False
            cls.settings.setValue(key, val)

    @staticmethod
    def makeLogger(deleteOldLogs=False, fileLogLvl=None):
        # send critical log messages to stdout
        logStream = logging.StreamHandler()
        logStream.setLevel(Core.stdoutLogLvl)
        streamFormatter = logging.Formatter("<%(name)s> %(levelname)s: %(message)s")
        logStream.setFormatter(streamFormatter)
        log = logging.getLogger("AVP")
        log.addHandler(logStream)

        if fileLogLvl is not None:
            # write log files as well!
            Core.logEnabled = True
            logFilename = os.path.join(Core.logDir, "avp_debug.log")
            libLogFilename = os.path.join(Core.logDir, "global_debug.log")

            if deleteOldLogs:
                for log_ in (logFilename, libLogFilename):
                    if os.path.exists(log_):
                        os.remove(log_)

            logFile = logging.FileHandler(logFilename, delay=True)
            logFile.setLevel(fileLogLvl)
            libLogFile = logging.FileHandler(libLogFilename, delay=True)
            libLogFile.setLevel(fileLogLvl)
            fileFormatter = logging.Formatter(
                "[%(asctime)s] %(threadName)-10.10s %(name)-23.23s %(levelname)s: "
                "%(message)s"
            )
            logFile.setFormatter(fileFormatter)
            libLogFile.setFormatter(fileFormatter)

            libLog = logging.getLogger()
            log.addHandler(logFile)
            libLog.addHandler(libLogFile)
            # lowest level must be explicitly set on the root Logger
            libLog.setLevel(0)

    @staticmethod
    def getConfigPath(dataDir=None):
        return (
            os.path.join(
                QtCore.QStandardPaths.writableLocation(
                    QtCore.QStandardPaths.StandardLocation.AppConfigLocation
                ),
                "audio-visualizer",
            )
            if dataDir is None
            else dataDir
        )



from pytestqt import qtbot
from pytest import fixture
from avp.toolkit.ffmpeg import checkFfmpegVersion
from . import command, imageDataSum, audioData, preFrameRender


@fixture
def coreWithWaveformComp(qtbot, command):
    """Fixture providing a Command object with Waveform component added"""
    command.settings.setValue("outputWidth", 1920)
    command.settings.setValue("outputHeight", 1080)
    command.core.insertComponent(0, command.core.moduleIndexFor("Waveform"), command)
    yield command.core


def test_comp_waveform_setColor(coreWithWaveformComp):
    comp = coreWithWaveformComp.selectedComponents[0]
    comp.page.lineEdit_color.setText("255,255,255")
    assert comp.color == (255, 255, 255)


def test_comp_waveform_previewRender(coreWithWaveformComp):
    comp = coreWithWaveformComp.selectedComponents[0]
    comp.page.lineEdit_color.setText("255,255,255")
    _, version = checkFfmpegVersion()
    if version > 6:
        # FFmpeg 8 has different colors from 6
        # TODO check version 7
        assert imageDataSum(comp.previewRender()) == 36114120
    else:
        assert imageDataSum(comp.previewRender()) == 37210620


def test_comp_waveform_renderFrame(coreWithWaveformComp, audioData):
    comp = coreWithWaveformComp.selectedComponents[0]
    comp.page.lineEdit_color.setText("255,255,255")
    preFrameRender(audioData, comp)
    image = comp.frameRender(0)
    comp.postFrameRender()
    assert imageDataSum(image) == 8331360
